#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import os

import pandas as pd
from elasticsearch import Elasticsearch

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

# Define test files and indices
ELASTICSEARCH_HOST = os.environ.get('ELASTICSEARCH_HOST') or 'localhost'

# Define client to use in tests
TEST_SUITE = os.environ.get('TEST_SUITE')
if TEST_SUITE == 'xpack':
    print('Running xpack tests requires SSL. Setting up SSL enabled client')
    certpath = os.path.join(os.path.dirname(__file__), '../../.ci/certs/ca.crt')
    print(certpath)
    ES_TEST_CLIENT = Elasticsearch(ELASTICSEARCH_HOST, use_ssl=True, verify_certs=True, ca_certs=certpath)
else:
    ES_TEST_CLIENT = Elasticsearch(ELASTICSEARCH_HOST)

FLIGHTS_INDEX_NAME = 'flights'
FLIGHTS_MAPPING = {"mappings": {
    "properties": {
        "AvgTicketPrice": {
            "type": "float"
        },
        "Cancelled": {
            "type": "boolean"
        },
        "Carrier": {
            "type": "keyword"
        },
        "Dest": {
            "type": "keyword"
        },
        "DestAirportID": {
            "type": "keyword"
        },
        "DestCityName": {
            "type": "keyword"
        },
        "DestCountry": {
            "type": "keyword"
        },
        "DestLocation": {
            "type": "geo_point"
        },
        "DestRegion": {
            "type": "keyword"
        },
        "DestWeather": {
            "type": "keyword"
        },
        "DistanceKilometers": {
            "type": "float"
        },
        "DistanceMiles": {
            "type": "float"
        },
        "FlightDelay": {
            "type": "boolean"
        },
        "FlightDelayMin": {
            "type": "integer"
        },
        "FlightDelayType": {
            "type": "keyword"
        },
        "FlightNum": {
            "type": "keyword"
        },
        "FlightTimeHour": {
            "type": "float"
        },
        "FlightTimeMin": {
            "type": "float"
        },
        "Origin": {
            "type": "keyword"
        },
        "OriginAirportID": {
            "type": "keyword"
        },
        "OriginCityName": {
            "type": "keyword"
        },
        "OriginCountry": {
            "type": "keyword"
        },
        "OriginLocation": {
            "type": "geo_point"
        },
        "OriginRegion": {
            "type": "keyword"
        },
        "OriginWeather": {
            "type": "keyword"
        },
        "dayOfWeek": {
            "type": "integer"
        },
        "timestamp": {
            "type": "date",
            "format": "strict_date_hour_minute_second"
        }
    }
}}
FLIGHTS_FILE_NAME = ROOT_DIR + '/flights.json.gz'
FLIGHTS_DF_FILE_NAME = ROOT_DIR + '/flights_df.json.gz'

FLIGHTS_SMALL_INDEX_NAME = 'flights_small'
FLIGHTS_SMALL_MAPPING = FLIGHTS_MAPPING
FLIGHTS_SMALL_FILE_NAME = ROOT_DIR + '/flights_small.json.gz'

ECOMMERCE_INDEX_NAME = 'ecommerce'
ECOMMERCE_MAPPING = {"mappings": {
    "properties": {
        "category": {
            "type": "text",
            "fields": {
                "keyword": {
                    "type": "keyword"
                }
            }
        },
        "currency": {
            "type": "keyword"
        },
        "customer_birth_date": {
            "type": "date"
        },
        "customer_first_name": {
            "type": "text",
            "fields": {
                "keyword": {
                    "type": "keyword",
                    "ignore_above": 256
                }
            }
        },
        "customer_full_name": {
            "type": "text",
            "fields": {
                "keyword": {
                    "type": "keyword",
                    "ignore_above": 256
                }
            }
        },
        "customer_gender": {
            "type": "text"
        },
        "customer_id": {
            "type": "keyword"
        },
        "customer_last_name": {
            "type": "text",
            "fields": {
                "keyword": {
                    "type": "keyword",
                    "ignore_above": 256
                }
            }
        },
        "customer_phone": {
            "type": "keyword"
        },
        "day_of_week": {
            "type": "keyword"
        },
        "day_of_week_i": {
            "type": "integer"
        },
        "email": {
            "type": "keyword"
        },
        "geoip": {
            "properties": {
                "city_name": {
                    "type": "keyword"
                },
                "continent_name": {
                    "type": "keyword"
                },
                "country_iso_code": {
                    "type": "keyword"
                },
                "location": {
                    "type": "geo_point"
                },
                "region_name": {
                    "type": "keyword"
                }
            }
        },
        "manufacturer": {
            "type": "text",
            "fields": {
                "keyword": {
                    "type": "keyword"
                }
            }
        },
        "order_date": {
            "type": "date"
        },
        "order_id": {
            "type": "keyword"
        },
        "products": {
            "properties": {
                "_id": {
                    "type": "text",
                    "fields": {
                        "keyword": {
                            "type": "keyword",
                            "ignore_above": 256
                        }
                    }
                },
                "base_price": {
                    "type": "half_float"
                },
                "base_unit_price": {
                    "type": "half_float"
                },
                "category": {
                    "type": "text",
                    "fields": {
                        "keyword": {
                            "type": "keyword"
                        }
                    }
                },
                "created_on": {
                    "type": "date"
                },
                "discount_amount": {
                    "type": "half_float"
                },
                "discount_percentage": {
                    "type": "half_float"
                },
                "manufacturer": {
                    "type": "text",
                    "fields": {
                        "keyword": {
                            "type": "keyword"
                        }
                    }
                },
                "min_price": {
                    "type": "half_float"
                },
                "price": {
                    "type": "half_float"
                },
                "product_id": {
                    "type": "long"
                },
                "product_name": {
                    "type": "text",
                    "fields": {
                        "keyword": {
                            "type": "keyword"
                        }
                    },
                    "analyzer": "english"
                },
                "quantity": {
                    "type": "integer"
                },
                "sku": {
                    "type": "keyword"
                },
                "tax_amount": {
                    "type": "half_float"
                },
                "taxful_price": {
                    "type": "half_float"
                },
                "taxless_price": {
                    "type": "half_float"
                },
                "unit_discount_amount": {
                    "type": "half_float"
                }
            }
        },
        "sku": {
            "type": "keyword"
        },
        "taxful_total_price": {
            "type": "float"
        },
        "taxless_total_price": {
            "type": "float"
        },
        "total_quantity": {
            "type": "integer"
        },
        "total_unique_products": {
            "type": "integer"
        },
        "type": {
            "type": "keyword"
        },
        "user": {
            "type": "keyword"
        }
    }
}}
ECOMMERCE_FILE_NAME = ROOT_DIR + '/ecommerce.json.gz'
ECOMMERCE_DF_FILE_NAME = ROOT_DIR + '/ecommerce_df.json.gz'

TEST_MAPPING1 = {
    'mappings': {
        'properties': {
            'city': {
                'type': 'text',
                'fields': {
                    'raw': {
                        'type': 'keyword'
                    }
                }
            },
            'text': {
                'type': 'text',
                'fields': {
                    'english': {
                        'type': 'text',
                        'analyzer': 'english'
                    }
                }
            },
            'origin_location': {
                'properties': {
                    'lat': {
                        'type': 'text',
                        'index_prefixes': {},
                        'fields': {
                            'keyword': {
                                'type': 'keyword',
                                'ignore_above': 256
                            }
                        }
                    },
                    'lon': {
                        'type': 'text',
                        'fields': {
                            'keyword': {
                                'type': 'keyword',
                                'ignore_above': 256
                            }
                        }
                    }
                }
            },
            'maps-telemetry': {
                'properties': {
                    'attributesPerMap': {
                        'properties': {
                            'dataSourcesCount': {
                                'properties': {
                                    'avg': {
                                        'type': 'long'
                                    },
                                    'max': {
                                        'type': 'long'
                                    },
                                    'min': {
                                        'type': 'long'
                                    }
                                }
                            },
                            'emsVectorLayersCount': {
                                'dynamic': 'true',
                                'properties': {
                                    'france_departments': {
                                        'properties': {
                                            'avg': {
                                                'type': 'float'
                                            },
                                            'max': {
                                                'type': 'long'
                                            },
                                            'min': {
                                                'type': 'long'
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            },
            'type': {
                'type': 'keyword'
            },
            'name': {
                'type': 'text'
            },
            'user_name': {
                'type': 'keyword'
            },
            'email': {
                'type': 'keyword'
            },
            'content': {
                'type': 'text'
            },
            'tweeted_at': {
                'type': 'date'
            },
            'dest_location': {
                'type': 'geo_point'
            },
            'my_join_field': {
                'type': 'join',
                'relations': {
                    'question': ['answer', 'comment'],
                    'answer': 'vote'
                }
            }
        }
    }
}

TEST_MAPPING1_INDEX_NAME = 'mapping1'

TEST_MAPPING1_EXPECTED = {
    'city': 'text',
    'city.raw': 'keyword',
    'content': 'text',
    'dest_location': 'geo_point',
    'email': 'keyword',
    'maps-telemetry.attributesPerMap.dataSourcesCount.avg': 'long',
    'maps-telemetry.attributesPerMap.dataSourcesCount.max': 'long',
    'maps-telemetry.attributesPerMap.dataSourcesCount.min': 'long',
    'maps-telemetry.attributesPerMap.emsVectorLayersCount.france_departments.avg': 'float',
    'maps-telemetry.attributesPerMap.emsVectorLayersCount.france_departments.max': 'long',
    'maps-telemetry.attributesPerMap.emsVectorLayersCount.france_departments.min': 'long',
    'my_join_field': 'join',
    'name': 'text',
    'origin_location.lat': 'text',
    'origin_location.lat.keyword': 'keyword',
    'origin_location.lon': 'text',
    'origin_location.lon.keyword': 'keyword',
    'text': 'text',
    'text.english': 'text',
    'tweeted_at': 'date',
    'type': 'keyword',
    'user_name': 'keyword'
}

TEST_MAPPING1_EXPECTED_DF = pd.DataFrame.from_dict(data=TEST_MAPPING1_EXPECTED, orient='index', columns=['es_dtype'])
TEST_MAPPING1_EXPECTED_SOURCE_FIELD_DF = TEST_MAPPING1_EXPECTED_DF.drop(index=['city.raw',
                                                                               'origin_location.lat.keyword',
                                                                               'origin_location.lon.keyword',
                                                                               'text.english'])
TEST_MAPPING1_EXPECTED_SOURCE_FIELD_COUNT = len(TEST_MAPPING1_EXPECTED_SOURCE_FIELD_DF.index)

TEST_NESTED_USER_GROUP_INDEX_NAME = 'nested_user_group'
TEST_NESTED_USER_GROUP_MAPPING = {
    'mappings': {
        'properties': {
            'group': {
                'type': 'keyword'
            },
            'user': {
                'properties': {
                    'first': {
                        'type': 'keyword'
                    },
                    'last': {
                        'type': 'keyword'
                    },
                    'address': {
                        'type': 'keyword'
                    }
                }
            }
        }
    }
}

TEST_NESTED_USER_GROUP_DOCS = [
    {'_index': TEST_NESTED_USER_GROUP_INDEX_NAME,
     '_source':
         {'group': 'amsterdam', 'user': [
             {'first': 'Manke', 'last': 'Nelis', 'address': ['Elandsgracht', 'Amsterdam']},
             {'first': 'Johnny', 'last': 'Jordaan', 'address': ['Elandsstraat', 'Amsterdam']}]}},
    {'_index': TEST_NESTED_USER_GROUP_INDEX_NAME,
     '_source':
         {'group': 'london', 'user': [
             {'first': 'Alice', 'last': 'Monkton'},
             {'first': 'Jimmy', 'last': 'White', 'address': ['London']}]}},
    {'_index': TEST_NESTED_USER_GROUP_INDEX_NAME,
     '_source': {'group': 'new york', 'user': [
         {'first': 'Bill', 'last': 'Jones'}]}}
]



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from elasticsearch import helpers
from elasticsearch.client import ClusterClient

from eland.tests import *

DATA_LIST = [
    (FLIGHTS_FILE_NAME, FLIGHTS_INDEX_NAME, FLIGHTS_MAPPING),
    (FLIGHTS_SMALL_FILE_NAME, FLIGHTS_SMALL_INDEX_NAME, FLIGHTS_MAPPING),
    (ECOMMERCE_FILE_NAME, ECOMMERCE_INDEX_NAME, ECOMMERCE_MAPPING)
]


def _setup_data(es):
    # Read json file and index records into Elasticsearch
    for data in DATA_LIST:
        json_file_name = data[0]
        index_name = data[1]
        mapping = data[2]

        # Delete index
        print("Deleting index:", index_name)
        es.indices.delete(index=index_name, ignore=[400, 404])
        print("Creating index:", index_name)
        es.indices.create(index=index_name, body=mapping)

        df = pd.read_json(json_file_name, lines=True)

        actions = []
        n = 0

        print("Adding", df.shape[0], "items to index:", index_name)
        for index, row in df.iterrows():
            values = row.to_dict()
            # make timestamp datetime 2018-01-01T12:09:35
            # values['timestamp'] = datetime.strptime(values['timestamp'], '%Y-%m-%dT%H:%M:%S')

            # Use integer as id field for repeatable results
            action = {'_index': index_name, '_source': values, '_id': str(n)}

            actions.append(action)

            n = n + 1

            if n % 10000 == 0:
                helpers.bulk(es, actions)
                actions = []

        helpers.bulk(es, actions)
        actions = []

        print("Done", index_name)


def _update_max_compilations_limit(es, limit="10000/1m"):
    print('Updating script.max_compilations_rate to ', limit)
    cluster_client = ClusterClient(es)
    body = {
        "transient": {
            "script.max_compilations_rate": limit
        }
    }
    cluster_client.put_settings(body=body)


def _setup_test_mappings(es):
    # Create a complex mapping containing many Elasticsearch features
    es.indices.delete(index=TEST_MAPPING1_INDEX_NAME, ignore=[400, 404])
    es.indices.create(index=TEST_MAPPING1_INDEX_NAME, body=TEST_MAPPING1)


def _setup_test_nested(es):
    es.indices.delete(index=TEST_NESTED_USER_GROUP_INDEX_NAME, ignore=[400, 404])
    es.indices.create(index=TEST_NESTED_USER_GROUP_INDEX_NAME, body=TEST_NESTED_USER_GROUP_MAPPING)

    helpers.bulk(es, TEST_NESTED_USER_GROUP_DOCS)


if __name__ == '__main__':
    # Create connection to Elasticsearch - use defaults
    print('Connecting to ES', ELASTICSEARCH_HOST)
    es = ES_TEST_CLIENT

    _setup_data(es)
    _setup_test_mappings(es)
    _setup_test_nested(es)
    _update_max_compilations_limit(es)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

import numpy as np

import eland as ed
from eland.tests import ES_TEST_CLIENT, ECOMMERCE_INDEX_NAME, FLIGHTS_INDEX_NAME
from eland.tests.common import TestData


class TestMetricSourceFields(TestData):

    def test_flights_all_metric_source_fields(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )
        pd_flights = self.pd_flights()

        ed_dtypes, ed_fields, es_date_formats = ed_field_mappings.metric_source_fields()
        pd_metric = pd_flights.select_dtypes(include=np.number)

        assert pd_metric.dtypes.to_list() == ed_dtypes
        assert pd_metric.columns.to_list() == ed_fields
        assert len(es_date_formats) == len(ed_dtypes)
        assert set(es_date_formats) == set([None])

    def test_flights_all_metric_source_fields_and_bool(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )
        pd_flights = self.pd_flights()

        ed_dtypes, ed_fields, es_date_formats = ed_field_mappings.metric_source_fields(include_bool=True)
        pd_metric = pd_flights.select_dtypes(include=[np.number, 'bool'])

        assert pd_metric.dtypes.to_list() == ed_dtypes
        assert pd_metric.columns.to_list() == ed_fields
        assert len(es_date_formats) == len(ed_dtypes)
        assert set(es_date_formats) == set([None])

    def test_flights_all_metric_source_fields_bool_and_timestamp(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )
        pd_flights = self.pd_flights()

        ed_dtypes, ed_fields, es_date_formats = ed_field_mappings.metric_source_fields(include_bool=True, include_timestamp=True)
        pd_metric = pd_flights.select_dtypes(include=[np.number, 'bool', 'datetime'])

        assert pd_metric.dtypes.to_list() == ed_dtypes
        assert pd_metric.columns.to_list() == ed_fields
        assert len(es_date_formats) == len(ed_dtypes)
        assert set(es_date_formats) == set({"strict_date_hour_minute_second", None}) # TODO - test position of date_format

    def test_ecommerce_selected_non_metric_source_fields(self):
        field_names = ['category', 'currency', 'customer_birth_date', 'customer_first_name', 'user']
        """
        Note: non of there are metric
        category                       object
        currency                       object
        customer_birth_date    datetime64[ns]
        customer_first_name            object
        user                           object
        """
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=ECOMMERCE_INDEX_NAME,
            display_names=field_names
        )
        pd_ecommerce = self.pd_ecommerce()[field_names]

        ed_dtypes, ed_fields, es_date_formats = ed_field_mappings.metric_source_fields()
        pd_metric = pd_ecommerce.select_dtypes(include=np.number)

        assert pd_metric.dtypes.to_list() == ed_dtypes
        assert pd_metric.columns.to_list() == ed_fields
        assert len(es_date_formats) == len(ed_dtypes)
        assert set(es_date_formats) == set()

    def test_ecommerce_selected_mixed_metric_source_fields(self):
        field_names = ['category', 'currency', 'customer_birth_date', 'customer_first_name', 'total_quantity', 'user']
        """
        Note: one is metric
        category                       object
        currency                       object
        customer_birth_date    datetime64[ns]
        customer_first_name            object
        total_quantity                 int64
        user                           object
        """
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=ECOMMERCE_INDEX_NAME,
            display_names=field_names
        )
        pd_ecommerce = self.pd_ecommerce()[field_names]

        ed_dtypes, ed_fields, es_date_formats = ed_field_mappings.metric_source_fields()
        pd_metric = pd_ecommerce.select_dtypes(include=np.number)
        assert len(es_date_formats) == len(ed_dtypes)
        assert set(es_date_formats) == set([None])

        assert pd_metric.dtypes.to_list() == ed_dtypes
        assert pd_metric.columns.to_list() == ed_fields

    def test_ecommerce_selected_all_metric_source_fields(self):
        field_names = ['total_quantity', 'taxful_total_price', 'taxless_total_price']
        """
        Note: all are metric
        total_quantity           int64
        taxful_total_price     float64
        taxless_total_price    float64
        """
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=ECOMMERCE_INDEX_NAME,
            display_names=field_names
        )
        pd_ecommerce = self.pd_ecommerce()[field_names]

        ed_dtypes, ed_fields, es_date_formats = ed_field_mappings.metric_source_fields()
        pd_metric = pd_ecommerce.select_dtypes(include=np.number)

        assert pd_metric.dtypes.to_list() == ed_dtypes
        assert pd_metric.columns.to_list() == ed_fields
        assert len(es_date_formats) == len(ed_dtypes)
        assert set(es_date_formats) == set([None])



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
import pytest

import eland as ed
from eland.tests import ES_TEST_CLIENT, ECOMMERCE_INDEX_NAME
from eland.tests.common import TestData


class TestAggregatables(TestData):

    @pytest.mark.filterwarnings("ignore:Aggregations not supported")
    def test_ecommerce_all_aggregatables(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=ECOMMERCE_INDEX_NAME
        )

        aggregatables = ed_field_mappings.aggregatable_field_names()

        expected = {'category.keyword': 'category',
                    'currency': 'currency',
                    'customer_birth_date': 'customer_birth_date',
                    'customer_first_name.keyword': 'customer_first_name',
                    'customer_full_name.keyword': 'customer_full_name',
                    'customer_id': 'customer_id',
                    'customer_last_name.keyword': 'customer_last_name',
                    'customer_phone': 'customer_phone',
                    'day_of_week': 'day_of_week',
                    'day_of_week_i': 'day_of_week_i',
                    'email': 'email',
                    'geoip.city_name': 'geoip.city_name',
                    'geoip.continent_name': 'geoip.continent_name',
                    'geoip.country_iso_code': 'geoip.country_iso_code',
                    'geoip.location': 'geoip.location',
                    'geoip.region_name': 'geoip.region_name',
                    'manufacturer.keyword': 'manufacturer',
                    'order_date': 'order_date',
                    'order_id': 'order_id',
                    'products._id.keyword': 'products._id',
                    'products.base_price': 'products.base_price',
                    'products.base_unit_price': 'products.base_unit_price',
                    'products.category.keyword': 'products.category',
                    'products.created_on': 'products.created_on',
                    'products.discount_amount': 'products.discount_amount',
                    'products.discount_percentage': 'products.discount_percentage',
                    'products.manufacturer.keyword': 'products.manufacturer',
                    'products.min_price': 'products.min_price',
                    'products.price': 'products.price',
                    'products.product_id': 'products.product_id',
                    'products.product_name.keyword': 'products.product_name',
                    'products.quantity': 'products.quantity',
                    'products.sku': 'products.sku',
                    'products.tax_amount': 'products.tax_amount',
                    'products.taxful_price': 'products.taxful_price',
                    'products.taxless_price': 'products.taxless_price',
                    'products.unit_discount_amount': 'products.unit_discount_amount',
                    'sku': 'sku',
                    'taxful_total_price': 'taxful_total_price',
                    'taxless_total_price': 'taxless_total_price',
                    'total_quantity': 'total_quantity',
                    'total_unique_products': 'total_unique_products',
                    'type': 'type',
                    'user': 'user'}

        assert expected == aggregatables

    def test_ecommerce_selected_aggregatables(self):
        expected = {'category.keyword': 'category',
                    'currency': 'currency',
                    'customer_birth_date': 'customer_birth_date',
                    'customer_first_name.keyword': 'customer_first_name',
                    'type': 'type', 'user': 'user'}

        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=ECOMMERCE_INDEX_NAME,
            display_names=expected.values()
        )

        aggregatables = ed_field_mappings.aggregatable_field_names()

        assert expected == aggregatables

    def test_ecommerce_single_aggregatable_field(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=ECOMMERCE_INDEX_NAME
        )

        assert 'user' == ed_field_mappings.aggregatable_field_name('user')

    def test_ecommerce_single_keyword_aggregatable_field(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=ECOMMERCE_INDEX_NAME
        )

        assert 'customer_first_name.keyword' == ed_field_mappings.aggregatable_field_name('customer_first_name')

    def test_ecommerce_single_non_existant_field(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=ECOMMERCE_INDEX_NAME
        )

        with pytest.raises(KeyError):
            aggregatable = ed_field_mappings.aggregatable_field_name('non_existant')

    @pytest.mark.filterwarnings("ignore:Aggregations not supported")
    def test_ecommerce_single_non_aggregatable_field(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=ECOMMERCE_INDEX_NAME
        )

        assert None == ed_field_mappings.aggregatable_field_name('customer_gender')



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from eland.tests.common import TestData


class TestDataFrameGet(TestData):

    def test_get_one_attribute(self):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_get0 = ed_flights.get('Carrier')
        pd_get0 = pd_flights.get('Carrier')

        print(ed_get0, type(ed_get0))
        print(pd_get0, type(pd_get0))



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.
import copy
import warnings
from collections import OrderedDict

import pandas as pd
from pandas.core.dtypes.common import is_bool_dtype, is_datetime_or_timedelta_dtype

from eland import Index, SortOrder, DEFAULT_CSV_BATCH_OUTPUT_SIZE, DEFAULT_ES_MAX_RESULT_WINDOW, \
    elasticsearch_date_to_pandas_date
from eland import Query
from eland.actions import SortFieldAction
from eland.tasks import HeadTask, TailTask, BooleanFilterTask, ArithmeticOpFieldsTask, QueryTermsTask, \
    QueryIdsTask, SizeTask


class Operations:
    """
    A collector of the queries and selectors we apply to queries to return the appropriate results.

    For example,
        - a list of the field_names in the DataFrame (a subset of field_names in the index)
        - a size limit on the results (e.g. for head(n=5))
        - a query to filter the results (e.g. df.A > 10)

    This is maintained as a 'task graph' (inspired by dask)
    (see https://docs.dask.org/en/latest/spec.html)
    """

    def __init__(self, tasks=None, arithmetic_op_fields_task=None):
        if tasks is None:
            self._tasks = []
        else:
            self._tasks = tasks
        self._arithmetic_op_fields_task = arithmetic_op_fields_task

    def __constructor__(self, *args, **kwargs):
        return type(self)(*args, **kwargs)

    def copy(self):
        return self.__constructor__(tasks=copy.deepcopy(self._tasks),
                                    arithmetic_op_fields_task=copy.deepcopy(self._arithmetic_op_fields_task))

    def head(self, index, n):
        # Add a task that is an ascending sort with size=n
        task = HeadTask(index.sort_field, n)
        self._tasks.append(task)

    def tail(self, index, n):
        # Add a task that is descending sort with size=n
        task = TailTask(index.sort_field, n)
        self._tasks.append(task)

    def arithmetic_op_fields(self, display_name, arithmetic_series):
        if self._arithmetic_op_fields_task is None:
            self._arithmetic_op_fields_task = ArithmeticOpFieldsTask(display_name, arithmetic_series)
        else:
            self._arithmetic_op_fields_task.update(display_name, arithmetic_series)

    def get_arithmetic_op_fields(self):
        # get an ArithmeticOpFieldsTask if it exists
        return self._arithmetic_op_fields_task

    def __repr__(self):
        return repr(self._tasks)

    def count(self, query_compiler):
        query_params, post_processing = self._resolve_tasks(query_compiler)

        # Elasticsearch _count is very efficient and so used to return results here. This means that
        # data frames that have restricted size or sort params will not return valid results
        # (_count doesn't support size).
        # Longer term we may fall back to pandas, but this may result in loading all index into memory.
        if self._size(query_params, post_processing) is not None:
            raise NotImplementedError("Requesting count with additional query and processing parameters "
                                      "not supported {0} {1}"
                                      .format(query_params, post_processing))

        # Only return requested field_names
        fields = query_compiler.get_field_names(include_scripted_fields=False)

        counts = OrderedDict()
        for field in fields:
            body = Query(query_params['query'])
            body.exists(field, must=True)

            field_exists_count = query_compiler._client.count(index=query_compiler._index_pattern,
                                                              body=body.to_count_body())
            counts[field] = field_exists_count

        return pd.Series(data=counts, index=fields)

    def mean(self, query_compiler, numeric_only=True):
        return self._metric_aggs(query_compiler, 'avg', numeric_only=numeric_only)

    def sum(self, query_compiler, numeric_only=True):
        return self._metric_aggs(query_compiler, 'sum', numeric_only=numeric_only)

    def max(self, query_compiler, numeric_only=True):
        return self._metric_aggs(query_compiler, 'max', numeric_only=numeric_only)

    def min(self, query_compiler, numeric_only=True):
        return self._metric_aggs(query_compiler, 'min', numeric_only=numeric_only)

    def nunique(self, query_compiler):
        return self._metric_aggs(query_compiler, 'cardinality', field_types='aggregatable')

    def value_counts(self, query_compiler, es_size):
        return self._terms_aggs(query_compiler, 'terms', es_size)

    def hist(self, query_compiler, bins):
        return self._hist_aggs(query_compiler, bins)

    def _metric_aggs(self, query_compiler, func, field_types=None, numeric_only=None):
        """
        Parameters
        ----------
        field_types: str, default None
            if `aggregatable` use only field_names whose fields in elasticseach are aggregatable.
            If `None`, use only numeric fields.

        Returns
        -------
        pandas.Series
            Series containing results of `func` applied to the field_name(s)
        """
        query_params, post_processing = self._resolve_tasks(query_compiler)

        size = self._size(query_params, post_processing)
        if size is not None:
            raise NotImplementedError("Can not count field matches if size is set {}".format(size))

        body = Query(query_params['query'])

        results = OrderedDict()

        # some metrics aggs (including cardinality) work on all aggregatable fields
        # therefore we include an optional all parameter on operations
        # that call _metric_aggs
        if field_types == 'aggregatable':
            aggregatable_field_names = query_compiler._mappings.aggregatable_field_names()

            for field in aggregatable_field_names.keys():
                body.metric_aggs(field, func, field)

            response = query_compiler._client.search(
                index=query_compiler._index_pattern,
                size=0,
                body=body.to_search_body())

            # Results are of the form
            # "aggregations" : {
            #   "customer_full_name.keyword" : {
            #     "value" : 10
            #   }
            # }

            # map aggregatable (e.g. x.keyword) to field_name
            for key, value in aggregatable_field_names.items():
                results[value] = response['aggregations'][key]['value']
        else:
            if numeric_only:
                pd_dtypes, source_fields, date_formats = query_compiler._mappings.metric_source_fields(
                    include_bool=True)
            else:
                # The only non-numerics we support are bool and timestamps currently
                # strings are not supported by metric aggs in ES
                # TODO - sum isn't supported for Timestamp in pandas - although ES does attempt to do it
                pd_dtypes, source_fields, date_formats = query_compiler._mappings.metric_source_fields(
                    include_bool=True,
                    include_timestamp=True)

            for field in source_fields:
                body.metric_aggs(field, func, field)

            response = query_compiler._client.search(
                index=query_compiler._index_pattern,
                size=0,
                body=body.to_search_body())

            # Results are of the form
            # "aggregations" : {
            #   "AvgTicketPrice" : {
            #     "value" : 628.2536888148849
            #   },
            #   "timestamp": {
            #     "value": 1.5165624455644382E12,
            #     "value_as_string": "2018-01-21T19:20:45.564Z"
            #   }
            # }
            for pd_dtype, field, date_format in zip(pd_dtypes, source_fields, date_formats):
                if is_datetime_or_timedelta_dtype(pd_dtype):
                    results[field] = elasticsearch_date_to_pandas_date(
                        response['aggregations'][field]['value_as_string'],
                        date_format
                    )
                else:
                    results[field] = response['aggregations'][field]['value']

        # Return single value if this is a series
        # if len(numeric_source_fields) == 1:
        #    return np.float64(results[numeric_source_fields[0]])
        s = pd.Series(data=results, index=results.keys())

        return s

    def _terms_aggs(self, query_compiler, func, es_size=None):
        """
        Parameters
        ----------
        es_size: int, default None
            Parameter used by Series.value_counts()

        Returns
        -------
        pandas.Series
            Series containing results of `func` applied to the field_name(s)
        """
        query_params, post_processing = self._resolve_tasks(query_compiler)

        size = self._size(query_params, post_processing)
        if size is not None:
            raise NotImplementedError("Can not count field matches if size is set {}".format(size))

        # Get just aggregatable field_names
        aggregatable_field_names = query_compiler._mappings.aggregatable_field_names()

        body = Query(query_params['query'])

        for field in aggregatable_field_names.keys():
            body.terms_aggs(field, func, field, es_size=es_size)

        response = query_compiler._client.search(
            index=query_compiler._index_pattern,
            size=0,
            body=body.to_search_body())

        results = OrderedDict()

        for key in aggregatable_field_names.keys():
            # key is aggregatable field, value is label
            # e.g. key=category.keyword, value=category
            for bucket in response['aggregations'][key]['buckets']:
                results[bucket['key']] = bucket['doc_count']

        try:
            # get first value in dict (key is .keyword)
            name = list(aggregatable_field_names.values())[0]
        except IndexError:
            name = None

        s = pd.Series(data=results, index=results.keys(), name=name)

        return s

    def _hist_aggs(self, query_compiler, num_bins):
        # Get histogram bins and weights for numeric field_names
        query_params, post_processing = self._resolve_tasks(query_compiler)

        size = self._size(query_params, post_processing)
        if size is not None:
            raise NotImplementedError("Can not count field matches if size is set {}".format(size))

        numeric_source_fields = query_compiler._mappings.numeric_source_fields()

        body = Query(query_params['query'])

        min_aggs = self._metric_aggs(query_compiler, 'min', numeric_only=True)
        max_aggs = self._metric_aggs(query_compiler, 'max', numeric_only=True)

        for field in numeric_source_fields:
            body.hist_aggs(field, field, min_aggs, max_aggs, num_bins)

        response = query_compiler._client.search(
            index=query_compiler._index_pattern,
            size=0,
            body=body.to_search_body())
        # results are like
        # "aggregations" : {
        #     "DistanceKilometers" : {
        #       "buckets" : [
        #         {
        #           "key" : 0.0,
        #           "doc_count" : 2956
        #         },
        #         {
        #           "key" : 1988.1482421875,
        #           "doc_count" : 768
        #         },
        #         ...

        bins = OrderedDict()
        weights = OrderedDict()

        # There is one more bin that weights
        # len(bins) = len(weights) + 1

        # bins = [  0.  36.  72. 108. 144. 180. 216. 252. 288. 324. 360.]
        # len(bins) == 11
        # weights = [10066.,   263.,   386.,   264.,   273.,   390.,   324.,   438.,   261.,   394.]
        # len(weights) == 10

        # ES returns
        # weights = [10066.,   263.,   386.,   264.,   273.,   390.,   324.,   438.,   261.,   252.,    142.]
        # So sum last 2 buckets
        for field in numeric_source_fields:

            # in case of series let plotting.ed_hist_series thrown an exception
            if not response.get('aggregations'):
                continue

            # in case of dataframe, throw warning that field is excluded
            if not response['aggregations'].get(field):
                warnings.warn("{} has no meaningful histogram interval and will be excluded. "
                              "All values 0."
                              .format(field),
                              UserWarning)
                continue

            buckets = response['aggregations'][field]['buckets']

            bins[field] = []
            weights[field] = []

            for bucket in buckets:
                bins[field].append(bucket['key'])

                if bucket == buckets[-1]:
                    weights[field][-1] += bucket['doc_count']
                else:
                    weights[field].append(bucket['doc_count'])

        df_bins = pd.DataFrame(data=bins)
        df_weights = pd.DataFrame(data=weights)

        return df_bins, df_weights

    @staticmethod
    def _map_pd_aggs_to_es_aggs(pd_aggs):
        """
        Args:
            pd_aggs - list of pandas aggs (e.g. ['mad', 'min', 'std'] etc.)

        Returns:
            ed_aggs - list of corresponding es_aggs (e.g. ['median_absolute_deviation', 'min', 'std'] etc.)

        Pandas supports a lot of options here, and these options generally work on text and numerics in pandas.
        Elasticsearch has metric aggs and terms aggs so will have different behaviour.

        Pandas aggs that return field_names (as opposed to transformed rows):

        all
        any
        count
        mad
        max
        mean
        median
        min
        mode
        quantile
        rank
        sem
        skew
        sum
        std
        var
        nunique
        """
        ed_aggs = []
        for pd_agg in pd_aggs:
            if pd_agg == 'count':
                ed_aggs.append('count')
            elif pd_agg == 'mad':
                ed_aggs.append('median_absolute_deviation')
            elif pd_agg == 'max':
                ed_aggs.append('max')
            elif pd_agg == 'mean':
                ed_aggs.append('avg')
            elif pd_agg == 'median':
                ed_aggs.append(('percentiles', '50.0'))
            elif pd_agg == 'min':
                ed_aggs.append('min')
            elif pd_agg == 'mode':
                # We could do this via top term
                raise NotImplementedError(pd_agg, " not currently implemented")
            elif pd_agg == 'quantile':
                # TODO
                raise NotImplementedError(pd_agg, " not currently implemented")
            elif pd_agg == 'rank':
                # TODO
                raise NotImplementedError(pd_agg, " not currently implemented")
            elif pd_agg == 'sem':
                # TODO
                raise NotImplementedError(pd_agg, " not currently implemented")
            elif pd_agg == 'sum':
                ed_aggs.append('sum')
            elif pd_agg == 'std':
                ed_aggs.append(('extended_stats', 'std_deviation'))
            elif pd_agg == 'var':
                ed_aggs.append(('extended_stats', 'variance'))
            else:
                raise NotImplementedError(pd_agg, " not currently implemented")

        # TODO - we can optimise extended_stats here as if we have 'count' and 'std' extended_stats would
        #   return both in one call

        return ed_aggs

    def aggs(self, query_compiler, pd_aggs):
        query_params, post_processing = self._resolve_tasks(query_compiler)

        size = self._size(query_params, post_processing)
        if size is not None:
            raise NotImplementedError("Can not count field matches if size is set {}".format(size))

        field_names = query_compiler.get_field_names(include_scripted_fields=False)

        body = Query(query_params['query'])

        # convert pandas aggs to ES equivalent
        es_aggs = self._map_pd_aggs_to_es_aggs(pd_aggs)

        for field in field_names:
            for es_agg in es_aggs:
                # If we have multiple 'extended_stats' etc. here we simply NOOP on 2nd call
                if isinstance(es_agg, tuple):
                    body.metric_aggs(es_agg[0] + '_' + field, es_agg[0], field)
                else:
                    body.metric_aggs(es_agg + '_' + field, es_agg, field)

        response = query_compiler._client.search(
            index=query_compiler._index_pattern,
            size=0,
            body=body.to_search_body())

        """
        Results are like (for 'sum', 'min')

             AvgTicketPrice  DistanceKilometers  DistanceMiles  FlightDelayMin
        sum    8.204365e+06        9.261629e+07   5.754909e+07          618150
        min    1.000205e+02        0.000000e+00   0.000000e+00               0
        """
        results = OrderedDict()

        for field in field_names:
            values = list()
            for es_agg in es_aggs:
                if isinstance(es_agg, tuple):
                    values.append(response['aggregations'][es_agg[0] + '_' + field][es_agg[1]])
                else:
                    values.append(response['aggregations'][es_agg + '_' + field]['value'])

            results[field] = values

        df = pd.DataFrame(data=results, index=pd_aggs)

        return df

    def describe(self, query_compiler):
        query_params, post_processing = self._resolve_tasks(query_compiler)

        size = self._size(query_params, post_processing)
        if size is not None:
            raise NotImplementedError("Can not count field matches if size is set {}".format(size))

        numeric_source_fields = query_compiler._mappings.numeric_source_fields()

        # for each field we compute:
        # count, mean, std, min, 25%, 50%, 75%, max
        body = Query(query_params['query'])

        for field in numeric_source_fields:
            body.metric_aggs('extended_stats_' + field, 'extended_stats', field)
            body.metric_aggs('percentiles_' + field, 'percentiles', field)

        response = query_compiler._client.search(
            index=query_compiler._index_pattern,
            size=0,
            body=body.to_search_body())

        results = OrderedDict()

        for field in numeric_source_fields:
            values = list()
            values.append(response['aggregations']['extended_stats_' + field]['count'])
            values.append(response['aggregations']['extended_stats_' + field]['avg'])
            values.append(response['aggregations']['extended_stats_' + field]['std_deviation'])
            values.append(response['aggregations']['extended_stats_' + field]['min'])
            values.append(response['aggregations']['percentiles_' + field]['values']['25.0'])
            values.append(response['aggregations']['percentiles_' + field]['values']['50.0'])
            values.append(response['aggregations']['percentiles_' + field]['values']['75.0'])
            values.append(response['aggregations']['extended_stats_' + field]['max'])

            # if not None
            if values.count(None) < len(values):
                results[field] = values

        df = pd.DataFrame(data=results, index=['count', 'mean', 'std', 'min', '25%', '50%', '75%', 'max'])

        return df

    def to_pandas(self, query_compiler, show_progress=False):
        class PandasDataFrameCollector:
            def __init__(self, show_progress):
                self._df = None
                self._show_progress = show_progress

            def collect(self, df):
                # This collector does not batch data on output. Therefore, batch_size is fixed to None and this method
                # is only called once.
                if self._df is not None:
                    raise RuntimeError("Logic error in execution, this method must only be called once for this"
                                       "collector - batch_size == None")
                self._df = df

            @staticmethod
            def batch_size():
                # Do not change (see notes on collect)
                return None

            @property
            def show_progress(self):
                return self._show_progress

        collector = PandasDataFrameCollector(show_progress)

        self._es_results(query_compiler, collector)

        return collector._df

    def to_csv(self, query_compiler, show_progress=False, **kwargs):
        class PandasToCSVCollector:
            def __init__(self, show_progress, **args):
                self._args = args
                self._show_progress = show_progress
                self._ret = None
                self._first_time = True

            def collect(self, df):
                # If this is the first time we collect results, then write header, otherwise don't write header
                # and append results
                if self._first_time:
                    self._first_time = False
                    df.to_csv(**self._args)
                else:
                    # Don't write header, and change mode to append
                    self._args['header'] = False
                    self._args['mode'] = 'a'
                    df.to_csv(**self._args)

            @staticmethod
            def batch_size():
                # By default read n docs and then dump to csv
                batch_size = DEFAULT_CSV_BATCH_OUTPUT_SIZE
                return batch_size

            @property
            def show_progress(self):
                return self._show_progress

        collector = PandasToCSVCollector(show_progress, **kwargs)

        self._es_results(query_compiler, collector)

        return collector._ret

    def _es_results(self, query_compiler, collector):
        query_params, post_processing = self._resolve_tasks(query_compiler)

        size, sort_params = Operations._query_params_to_size_and_sort(query_params)

        script_fields = query_params['query_script_fields']
        query = Query(query_params['query'])

        body = query.to_search_body()
        if script_fields is not None:
            body['script_fields'] = script_fields

        # Only return requested field_names
        _source = query_compiler.get_field_names(include_scripted_fields=False)
        if not _source:
            _source = False

        es_results = None

        # If size=None use scan not search - then post sort results when in df
        # If size>10000 use scan
        is_scan = False
        if size is not None and size <= DEFAULT_ES_MAX_RESULT_WINDOW:
            if size > 0:
                try:
                    es_results = query_compiler._client.search(
                        index=query_compiler._index_pattern,
                        size=size,
                        sort=sort_params,
                        body=body,
                        _source=_source)
                except Exception:
                    # Catch all ES errors and print debug (currently to stdout)
                    error = {
                        'index': query_compiler._index_pattern,
                        'size': size,
                        'sort': sort_params,
                        'body': body,
                        '_source': _source
                    }
                    print("Elasticsearch error:", error)
                    raise
        else:
            is_scan = True
            es_results = query_compiler._client.scan(
                index=query_compiler._index_pattern,
                query=body,
                _source=_source)
            # create post sort
            if sort_params is not None:
                post_processing.append(SortFieldAction(sort_params))

        if is_scan:
            while True:
                partial_result, df = query_compiler._es_results_to_pandas(es_results, collector.batch_size(),
                                                                          collector.show_progress)
                df = self._apply_df_post_processing(df, post_processing)
                collector.collect(df)
                if not partial_result:
                    break
        else:
            partial_result, df = query_compiler._es_results_to_pandas(es_results)
            df = self._apply_df_post_processing(df, post_processing)
            collector.collect(df)

    def index_count(self, query_compiler, field):
        # field is the index field so count values
        query_params, post_processing = self._resolve_tasks(query_compiler)

        size = self._size(query_params, post_processing)

        # Size is dictated by operations
        if size is not None:
            # TODO - this is not necessarily valid as the field may not exist in ALL these docs
            return size

        body = Query(query_params['query'])
        body.exists(field, must=True)

        return query_compiler._client.count(index=query_compiler._index_pattern, body=body.to_count_body())

    def _validate_index_operation(self, query_compiler, items):
        if not isinstance(items, list):
            raise TypeError("list item required - not {}".format(type(items)))

        # field is the index field so count values
        query_params, post_processing = self._resolve_tasks(query_compiler)

        size = self._size(query_params, post_processing)

        # Size is dictated by operations
        if size is not None:
            raise NotImplementedError("Can not count field matches if size is set {}".format(size))

        return query_params, post_processing

    def index_matches_count(self, query_compiler, field, items):
        query_params, post_processing = self._validate_index_operation(query_compiler, items)

        body = Query(query_params['query'])

        if field == Index.ID_INDEX_FIELD:
            body.ids(items, must=True)
        else:
            body.terms(field, items, must=True)

        return query_compiler._client.count(index=query_compiler._index_pattern, body=body.to_count_body())

    def drop_index_values(self, query_compiler, field, items):
        self._validate_index_operation(query_compiler, items)

        # Putting boolean queries together
        # i = 10
        # not i = 20
        # _id in [1,2,3]
        # _id not in [1,2,3]
        # a in ['a','b','c']
        # b not in ['a','b','c']
        # For now use term queries
        if field == Index.ID_INDEX_FIELD:
            task = QueryIdsTask(False, items)
        else:
            task = QueryTermsTask(False, field, items)
        self._tasks.append(task)

    @staticmethod
    def _query_params_to_size_and_sort(query_params):
        sort_params = None
        if query_params['query_sort_field'] and query_params['query_sort_order']:
            sort_params = query_params['query_sort_field'] + ":" + SortOrder.to_string(
                query_params['query_sort_order'])

        size = query_params['query_size']

        return size, sort_params

    @staticmethod
    def _count_post_processing(post_processing):
        size = None
        for action in post_processing:
            if isinstance(action, SizeTask):
                if size is None or action.size() < size:
                    size = action.size()

        return size

    @staticmethod
    def _apply_df_post_processing(df, post_processing):
        for action in post_processing:
            df = action.resolve_action(df)

        return df

    def _resolve_tasks(self, query_compiler):
        # We now try and combine all tasks into an Elasticsearch query
        # Some operations can be simply combined into a single query
        # other operations require pre-queries and then combinations
        # other operations require in-core post-processing of results
        query_params = {"query_sort_field": None,
                        "query_sort_order": None,
                        "query_size": None,
                        "query_fields": None,
                        "query_script_fields": None,
                        "query": Query()}

        post_processing = []

        for task in self._tasks:
            query_params, post_processing = task.resolve_task(query_params, post_processing, query_compiler)

        if self._arithmetic_op_fields_task is not None:
            query_params, post_processing = self._arithmetic_op_fields_task.resolve_task(query_params, post_processing,
                                                                                         query_compiler)

        return query_params, post_processing

    def _size(self, query_params, post_processing):
        # Shrink wrap code around checking if size parameter is set
        size = query_params['query_size']  # can be None

        pp_size = self._count_post_processing(post_processing)
        if pp_size is not None:
            if size is not None:
                size = min(size, pp_size)
            else:
                size = pp_size

        # This can return None
        return size

    def info_es(self, query_compiler, buf):
        buf.write("Operations:\n")
        buf.write(" tasks: {0}\n".format(self._tasks))

        query_params, post_processing = self._resolve_tasks(query_compiler)
        size, sort_params = Operations._query_params_to_size_and_sort(query_params)
        _source = query_compiler._mappings.get_field_names()

        script_fields = query_params['query_script_fields']
        query = Query(query_params['query'])
        body = query.to_search_body()
        if script_fields is not None:
            body['script_fields'] = script_fields

        buf.write(" size: {0}\n".format(size))
        buf.write(" sort_params: {0}\n".format(sort_params))
        buf.write(" _source: {0}\n".format(_source))
        buf.write(" body: {0}\n".format(body))
        buf.write(" post_processing: {0}\n".format(post_processing))

    def update_query(self, boolean_filter):
        task = BooleanFilterTask(boolean_filter)
        self._tasks.append(task)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from abc import ABC, abstractmethod

# -------------------------------------------------------------------------------------------------------------------- #
# PostProcessingActions                                                                                                #
# -------------------------------------------------------------------------------------------------------------------- #
from eland import SortOrder


class PostProcessingAction(ABC):
    def __init__(self, action_type):
        """
        Abstract class for postprocessing actions

        Parameters
        ----------
            action_type: str
                The action type (e.g. sort_index, head etc.)
        """
        self._action_type = action_type

    @property
    def type(self):
        return self._action_type

    @abstractmethod
    def resolve_action(self, df):
        pass

    @abstractmethod
    def __repr__(self):
        pass


class SortIndexAction(PostProcessingAction):
    def __init__(self):
        super().__init__("sort_index")

    def resolve_action(self, df):
        return df.sort_index()

    def __repr__(self):
        return "('{}')".format(self.type)


class HeadAction(PostProcessingAction):
    def __init__(self, count):
        super().__init__("head")

        self._count = count

    def resolve_action(self, df):
        return df.head(self._count)

    def __repr__(self):
        return "('{}': ('count': {}))".format(self.type, self._count)


class TailAction(PostProcessingAction):
    def __init__(self, count):
        super().__init__("tail")

        self._count = count

    def resolve_action(self, df):
        return df.tail(self._count)

    def __repr__(self):
        return "('{}': ('count': {}))".format(self.type, self._count)


class SortFieldAction(PostProcessingAction):
    def __init__(self, sort_params_string):
        super().__init__("sort_field")

        if sort_params_string is None:
            raise ValueError("Expected valid string")

        # Split string
        sort_params = sort_params_string.split(":")
        if len(sort_params) != 2:
            raise ValueError("Expected ES sort params string (e.g. _doc:desc). Got '{}'".format(sort_params_string))

        self._sort_field = sort_params[0]
        self._sort_order = SortOrder.from_string(sort_params[1])

    def resolve_action(self, df):
        if self._sort_order == SortOrder.ASC:
            return df.sort_values(self._sort_field, True)
        return df.sort_values(self._sort_field, False)

    def __repr__(self):
        return "('{}': ('sort_field': '{}', 'sort_order': {}))".format(self.type, self._sort_field, self._sort_order)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

import ast
import time

import pandas as pd
from pandas.util.testing import assert_frame_equal

import eland as ed
from eland.tests import ES_TEST_CLIENT
from eland.tests import FLIGHTS_INDEX_NAME
from eland.tests.common import ROOT_DIR
from eland.tests.common import TestData


class TestDataFrameToCSV(TestData):

    def test_to_csv_head(self):
        results_file = ROOT_DIR + '/dataframe/results/test_to_csv_head.csv'

        ed_flights = self.ed_flights().head()
        pd_flights = self.pd_flights().head()
        ed_flights.to_csv(results_file)
        # Converting back from csv is messy as pd_flights is created from a json file
        pd_from_csv = pd.read_csv(results_file, index_col=0, converters={
            'DestLocation': lambda x: ast.literal_eval(x),
            'OriginLocation': lambda x: ast.literal_eval(x)})
        pd_from_csv.index = pd_from_csv.index.map(str)
        pd_from_csv.timestamp = pd.to_datetime(pd_from_csv.timestamp)

        assert_frame_equal(pd_flights, pd_from_csv)

    def test_to_csv_full(self):
        results_file = ROOT_DIR + '/dataframe/results/test_to_csv_full.csv'

        # Test is slow as it's for the full dataset, but it is useful as it goes over 10000 docs
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_flights.to_csv(results_file)
        # Converting back from csv is messy as pd_flights is created from a json file
        pd_from_csv = pd.read_csv(results_file, index_col=0, converters={
            'DestLocation': lambda x: ast.literal_eval(x),
            'OriginLocation': lambda x: ast.literal_eval(x)})
        pd_from_csv.index = pd_from_csv.index.map(str)
        pd_from_csv.timestamp = pd.to_datetime(pd_from_csv.timestamp)

        assert_frame_equal(pd_flights, pd_from_csv)

        # Now read the csv to an index
        now_millis = int(round(time.time() * 1000))

        test_index = FLIGHTS_INDEX_NAME + '.' + str(now_millis)

        ed_flights_from_csv = ed.read_csv(results_file, ES_TEST_CLIENT, test_index, index_col=0, es_refresh=True,
                                          es_geo_points=['OriginLocation', 'DestLocation'],
                                          converters={
                                              'DestLocation': lambda x: ast.literal_eval(x),
                                              'OriginLocation': lambda x: ast.literal_eval(x)}
                                          )
        pd_flights_from_csv = ed.eland_to_pandas(ed_flights_from_csv)

        # TODO - there is a 'bug' where the Elasticsearch index returns data in a different order to the CSV
        print(ed_flights_from_csv.head())
        print(pd_flights_from_csv.head())

        # clean up index
        ES_TEST_CLIENT.indices.delete(test_index)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# Default number of rows displayed (different to pandas where ALL could be displayed)
import warnings
from enum import Enum
from typing import Union

import pandas as pd

DEFAULT_NUM_ROWS_DISPLAYED = 60

DEFAULT_CHUNK_SIZE = 10000
DEFAULT_CSV_BATCH_OUTPUT_SIZE = 10000
DEFAULT_PROGRESS_REPORTING_NUM_ROWS = 10000
DEFAULT_ES_MAX_RESULT_WINDOW = 10000  # index.max_result_window


def docstring_parameter(*sub):
    def dec(obj):
        obj.__doc__ = obj.__doc__.format(*sub)
        return obj

    return dec


class SortOrder(Enum):
    ASC = 0
    DESC = 1

    @staticmethod
    def reverse(order):
        if order == SortOrder.ASC:
            return SortOrder.DESC

        return SortOrder.ASC

    @staticmethod
    def to_string(order):
        if order == SortOrder.ASC:
            return "asc"

        return "desc"

    @staticmethod
    def from_string(order):
        if order == "asc":
            return SortOrder.ASC

        return SortOrder.DESC

def elasticsearch_date_to_pandas_date(value: Union[int, str], date_format: str) -> pd.Timestamp:
    """
    Given a specific Elasticsearch format for a date datatype, returns the
    'partial' `to_datetime` function to parse a given value in that format

    **Date Formats: https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping-date-format.html#built-in-date-formats

    Parameters
    ----------
    value: Union[int, str]
        The date value.
    date_format: str
        The Elasticsearch date format (ex. 'epoch_millis', 'epoch_second', etc.)

    Returns
    -------
    datetime: pd.Timestamp
        From https://www.elastic.co/guide/en/elasticsearch/reference/current/date.html
        Date formats can be customised, but if no format is specified then it uses the default:
        "strict_date_optional_time||epoch_millis"
        Therefore if no format is specified we assume either strict_date_optional_time
        or epoch_millis.
    """

    if date_format is None:
        try:
            value = int(value)
            return pd.to_datetime(value, unit='ms')
        except ValueError:
            return pd.to_datetime(value)
    elif date_format == "epoch_millis":
        return pd.to_datetime(value, unit='ms')
    elif date_format == "epoch_second":
        return pd.to_datetime(value, unit='s')
    elif date_format == "strict_date_optional_time":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S.%f%z", exact=False)
    elif date_format == "basic_date":
        return pd.to_datetime(value, format="%Y%m%d")
    elif date_format == "basic_date_time":
        return pd.to_datetime(value, format="%Y%m%dT%H%M%S.%f", exact=False)
    elif date_format == "basic_date_time_no_millis":
        return pd.to_datetime(value, format="%Y%m%dT%H%M%S%z")
    elif date_format == "basic_ordinal_date":
        return pd.to_datetime(value, format="%Y%j")
    elif date_format == "basic_ordinal_date_time":
        return pd.to_datetime(value, format="%Y%jT%H%M%S.%f%z", exact=False)
    elif date_format == "basic_ordinal_date_time_no_millis":
        return pd.to_datetime(value, format="%Y%jT%H%M%S%z")
    elif date_format == "basic_time":
        return pd.to_datetime(value, format="%H%M%S.%f%z", exact=False)
    elif date_format == "basic_time_no_millis":
        return pd.to_datetime(value, format="%H%M%S%z")
    elif date_format == "basic_t_time":
        return pd.to_datetime(value, format="T%H%M%S.%f%z", exact=False)
    elif date_format == "basic_t_time_no_millis":
        return pd.to_datetime(value, format="T%H%M%S%z")
    elif date_format == "basic_week_date":
        return pd.to_datetime(value, format="%GW%V%u")
    elif date_format == "basic_week_date_time":
        return pd.to_datetime(value, format="%GW%V%uT%H%M%S.%f%z", exact=False)
    elif date_format == "basic_week_date_time_no_millis":
        return pd.to_datetime(value, format="%GW%V%uT%H%M%S%z")
    elif date_format == "strict_date":
        return pd.to_datetime(value, format="%Y-%m-%d")
    elif date_format == "date":
        return pd.to_datetime(value, format="%Y-%m-%d")
    elif date_format == "strict_date_hour":
        return pd.to_datetime(value, format="%Y-%m-%dT%H")
    elif date_format == "date_hour":
        return pd.to_datetime(value, format="%Y-%m-%dT%H")
    elif date_format == "strict_date_hour_minute":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M")
    elif date_format == "date_hour_minute":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M")
    elif date_format == "strict_date_hour_minute_second":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S")
    elif date_format == "date_hour_minute_second":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S")
    elif date_format == "strict_date_hour_minute_second_fraction":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S.%f", exact=False)
    elif date_format == "date_hour_minute_second_fraction":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S.%f", exact=False)
    elif date_format == "strict_date_hour_minute_second_millis":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S.%f", exact=False)
    elif date_format == "date_hour_minute_second_millis":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S.%f", exact=False)
    elif date_format == "strict_date_time":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S.%f%z", exact=False)
    elif date_format == "date_time":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S.%f%z", exact=False)
    elif date_format == "strict_date_time_no_millis":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S%z")
    elif date_format == "date_time_no_millis":
        return pd.to_datetime(value, format="%Y-%m-%dT%H:%M:%S%z")
    elif date_format == "strict_hour":
        return pd.to_datetime(value, format="%H")
    elif date_format == "hour":
        return pd.to_datetime(value, format="%H")
    elif date_format == "strict_hour_minute":
        return pd.to_datetime(value, format="%H:%M")
    elif date_format == "hour_minute":
        return pd.to_datetime(value, format="%H:%M")
    elif date_format == "strict_hour_minute_second":
        return pd.to_datetime(value, format="%H:%M:%S")
    elif date_format == "hour_minute_second":
        return pd.to_datetime(value, format="%H:%M:%S")
    elif date_format == "strict_hour_minute_second_fraction":
        return pd.to_datetime(value, format="%H:%M:%S.%f", exact=False)
    elif date_format == "hour_minute_second_fraction":
        return pd.to_datetime(value, format="%H:%M:%S.%f", exact=False)
    elif date_format == "strict_hour_minute_second_millis":
        return pd.to_datetime(value, format="%H:%M:%S.%f", exact=False)
    elif date_format == "hour_minute_second_millis":
        return pd.to_datetime(value, format="%H:%M:%S.%f", exact=False)
    elif date_format == "strict_ordinal_date":
        return pd.to_datetime(value, format="%Y-%j")
    elif date_format == "ordinal_date":
        return pd.to_datetime(value, format="%Y-%j")
    elif date_format == "strict_ordinal_date_time":
        return pd.to_datetime(value, format="%Y-%jT%H:%M:%S.%f%z", exact=False)
    elif date_format == "ordinal_date_time":
        return pd.to_datetime(value, format="%Y-%jT%H:%M:%S.%f%z", exact=False)
    elif date_format == "strict_ordinal_date_time_no_millis":
        return pd.to_datetime(value, format="%Y-%jT%H:%M:%S%z")
    elif date_format == "ordinal_date_time_no_millis":
        return pd.to_datetime(value, format="%Y-%jT%H:%M:%S%z")
    elif date_format == "strict_time":
        return pd.to_datetime(value, format="%H:%M:%S.%f%z", exact=False)
    elif date_format == "time":
        return pd.to_datetime(value, format="%H:%M:%S.%f%z", exact=False)
    elif date_format == "strict_time_no_millis":
        return pd.to_datetime(value, format="%H:%M:%S%z")
    elif date_format == "time_no_millis":
        return pd.to_datetime(value, format="%H:%M:%S%z")
    elif date_format == "strict_t_time":
        return pd.to_datetime(value, format="T%H:%M:%S.%f%z", exact=False)
    elif date_format == "t_time":
        return pd.to_datetime(value, format="T%H:%M:%S.%f%z", exact=False)
    elif date_format == "strict_t_time_no_millis":
        return pd.to_datetime(value, format="T%H:%M:%S%z")
    elif date_format == "t_time_no_millis":
        return pd.to_datetime(value, format="T%H:%M:%S%z")
    elif date_format == "strict_week_date":
        return pd.to_datetime(value, format="%G-W%V-%u")
    elif date_format == "week_date":
        return pd.to_datetime(value, format="%G-W%V-%u")
    elif date_format == "strict_week_date_time":
        return pd.to_datetime(value, format="%G-W%V-%uT%H:%M:%S.%f%z", exact=False)
    elif date_format == "week_date_time":
        return pd.to_datetime(value, format="%G-W%V-%uT%H:%M:%S.%f%z", exact=False)
    elif date_format == "strict_week_date_time_no_millis":
        return pd.to_datetime(value, format="%G-W%V-%uT%H:%M:%S%z")
    elif date_format == "week_date_time_no_millis":
        return pd.to_datetime(value, format="%G-W%V-%uT%H:%M:%S%z")
    elif date_format == "strict_weekyear" or date_format == "weekyear":
        # TODO investigate if there is a way of converting this
        raise NotImplementedError("strict_weekyear is not implemented due to support in pandas")
        return pd.to_datetime(value, format="%G")
        # Not supported in pandas
        # ValueError: ISO year directive '%G' must be used with the ISO week directive '%V'
        # and a weekday directive '%A', '%a', '%w', or '%u'.
    elif date_format == "strict_weekyear_week" or date_format == "weekyear_week":
        # TODO investigate if there is a way of converting this
        raise NotImplementedError("strict_weekyear_week is not implemented due to support in pandas")
        return pd.to_datetime(value, format="%G-W%V")
        # Not supported in pandas
        # ValueError: ISO year directive '%G' must be used with the ISO week directive '%V'
        # and a weekday directive '%A', '%a', '%w', or '%u'.
    elif date_format == "strict_weekyear_week_day":
        return pd.to_datetime(value, format="%G-W%V-%u")
    elif date_format == "weekyear_week_day":
        return pd.to_datetime(value, format="%G-W%V-%u")
    elif date_format == "strict_year":
        return pd.to_datetime(value, format="%Y")
    elif date_format == "year":
        return pd.to_datetime(value, format="%Y")
    elif date_format == "strict_year_month":
        return pd.to_datetime(value, format="%Y-%m")
    elif date_format == "year_month":
        return pd.to_datetime(value, format="%Y-%m")
    elif date_format == "strict_year_month_day":
        return pd.to_datetime(value, format="%Y-%m-%d")
    elif date_format == "year_month_day":
        return pd.to_datetime(value, format="%Y-%m-%d")
    else:
        warnings.warn("The '{}' format is not explicitly supported."
                      "Using pandas.to_datetime(value) to parse value".format(date_format),
                      Warning)
        # TODO investigate how we could generate this just once for a bulk read.
        return pd.to_datetime(value)




#  Copyright 2019 Elasticsearch BV
# 
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
# 
#          http://www.apache.org/licenses/LICENSE-2.0
# 
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
from elasticsearch import Elasticsearch

import eland as ed
from eland.tests.common import TestData


class TestClientEq(TestData):

    def test_self_eq(self):
        es = Elasticsearch('localhost')

        client = ed.Client(es)

        assert client != es

        assert client == client

    def test_non_self_ne(self):
        es1 = Elasticsearch('localhost')
        es2 = Elasticsearch('localhost')

        client1 = ed.Client(es1)
        client2 = ed.Client(es2)

        assert client1 != client2



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

import pytest

import eland as ed
from eland.tests import ES_TEST_CLIENT
from eland.tests import FLIGHTS_INDEX_NAME


class TestDataFrameInit:

    def test_init(self):
        # Construct empty DataFrame (throws)
        with pytest.raises(ValueError):
            df = ed.DataFrame()

        # Construct invalid DataFrame (throws)
        with pytest.raises(ValueError):
            df = ed.DataFrame(client=ES_TEST_CLIENT)

        # Construct invalid DataFrame (throws)
        with pytest.raises(ValueError):
            df = ed.DataFrame(index_pattern=FLIGHTS_INDEX_NAME)

        # Good constructors
        df0 = ed.DataFrame(ES_TEST_CLIENT, FLIGHTS_INDEX_NAME)
        df1 = ed.DataFrame(client=ES_TEST_CLIENT, index_pattern=FLIGHTS_INDEX_NAME)

        qc = ed.QueryCompiler(client=ES_TEST_CLIENT, index_pattern=FLIGHTS_INDEX_NAME)
        df2 = ed.DataFrame(query_compiler=qc)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from pandas.util.testing import assert_almost_equal

from eland.tests.common import TestData


class TestSeriesMetrics(TestData):
    funcs = ['max', 'min', 'mean', 'sum']
    timestamp_funcs = ['max', 'min', 'mean']

    def test_flights_metrics(self):
        pd_flights = self.pd_flights()['AvgTicketPrice']
        ed_flights = self.ed_flights()['AvgTicketPrice']

        for func in self.funcs:
            pd_metric = getattr(pd_flights, func)()
            ed_metric = getattr(ed_flights, func)()
            assert_almost_equal(pd_metric, ed_metric, check_less_precise=True)

    def test_flights_timestamp(self):
        pd_flights = self.pd_flights()['timestamp']
        ed_flights = self.ed_flights()['timestamp']

        for func in self.timestamp_funcs:
            pd_metric = getattr(pd_flights, func)()
            ed_metric = getattr(ed_flights, func)()
            pd_metric = pd_metric.floor("S") # floor or pandas mean with have ns
            assert_almost_equal(pd_metric, ed_metric, check_less_precise=True)

    def test_ecommerce_selected_non_numeric_source_fields(self):
        # None of these are numeric
        column = 'category'

        ed_ecommerce = self.ed_ecommerce()[column]

        for func in self.funcs:
            ed_metric = getattr(ed_ecommerce, func)()
            assert ed_metric.empty

    def test_ecommerce_selected_all_numeric_source_fields(self):
        # All of these are numeric
        columns = ['total_quantity', 'taxful_total_price', 'taxless_total_price']

        for column in columns:
            pd_ecommerce = self.pd_ecommerce()[column]
            ed_ecommerce = self.ed_ecommerce()[column]

            for func in self.funcs:
                assert_almost_equal(getattr(pd_ecommerce, func)(), getattr(ed_ecommerce, func)(),
                                    check_less_precise=True)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
import pytest

from eland.tests.common import TestData, assert_pandas_eland_series_equal


class TestSeriesArithmetics(TestData):

    def test_invalid_add_num(self):
        with pytest.raises(TypeError):
            assert 2 + self.ed_ecommerce()['currency']

        with pytest.raises(TypeError):
            assert self.ed_ecommerce()['currency'] + 2

        with pytest.raises(TypeError):
            assert self.ed_ecommerce()['currency'] + self.ed_ecommerce()['total_quantity']

        with pytest.raises(TypeError):
            assert self.ed_ecommerce()['total_quantity'] + self.ed_ecommerce()['currency']

    def test_ser_add_ser(self):
        edadd = self.ed_ecommerce()['customer_first_name'] + self.ed_ecommerce()['customer_last_name']
        pdadd = self.pd_ecommerce()['customer_first_name'] + self.pd_ecommerce()['customer_last_name']

        assert_pandas_eland_series_equal(pdadd, edadd)

    def test_ser_add_str(self):
        edadd = self.ed_ecommerce()['customer_first_name'] + " is the first name."
        pdadd = self.pd_ecommerce()['customer_first_name'] + " is the first name."

        assert_pandas_eland_series_equal(pdadd, edadd)

    def test_frame_add_str(self):
        pdadd = self.pd_ecommerce()[['customer_first_name', 'customer_last_name']] + "_steve"
        print(pdadd.head())
        print(pdadd.columns)

    def test_str_add_ser(self):
        edadd = "The last name is: " + self.ed_ecommerce()['customer_last_name']
        pdadd = "The last name is: " + self.pd_ecommerce()['customer_last_name']

        assert_pandas_eland_series_equal(pdadd, edadd)

    def test_bad_str_add_ser(self):
        # TODO encode special characters better
        #      Elasticsearch accepts this, but it will cause problems
        edadd = " *" + self.ed_ecommerce()['customer_last_name']
        pdadd = " *" + self.pd_ecommerce()['customer_last_name']

        assert_pandas_eland_series_equal(pdadd, edadd)

    def test_ser_add_str_add_ser(self):
        pdadd = self.pd_ecommerce()['customer_first_name'] + " " + self.pd_ecommerce()['customer_last_name']
        edadd = self.ed_ecommerce()['customer_first_name'] + " " + self.ed_ecommerce()['customer_last_name']

        assert_pandas_eland_series_equal(pdadd, edadd)

    @pytest.mark.filterwarnings("ignore:Aggregations not supported")
    def test_non_aggregatable_add_str(self):
        with pytest.raises(ValueError):
            assert self.ed_ecommerce()['customer_gender'] + "is the gender"

    @pytest.mark.filterwarnings("ignore:Aggregations not supported")
    def teststr_add_non_aggregatable(self):
        with pytest.raises(ValueError):
            assert "The gender is: " + self.ed_ecommerce()['customer_gender']

    @pytest.mark.filterwarnings("ignore:Aggregations not supported")
    def test_non_aggregatable_add_aggregatable(self):
        with pytest.raises(ValueError):
            assert self.ed_ecommerce()['customer_gender'] + self.ed_ecommerce()['customer_first_name']

    @pytest.mark.filterwarnings("ignore:Aggregations not supported")
    def test_aggregatable_add_non_aggregatable(self):
        with pytest.raises(ValueError):
            assert self.ed_ecommerce()['customer_first_name'] + self.ed_ecommerce()['customer_gender']



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
from pandas.util.testing import assert_series_equal

import eland as ed
from eland.tests import ES_TEST_CLIENT, FLIGHTS_INDEX_NAME
from eland.tests.common import TestData


class TestDTypes(TestData):

    def test_all_fields(self):
        field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        pd_flights = self.pd_flights()

        assert_series_equal(pd_flights.dtypes, field_mappings.dtypes())

    def test_selected_fields(self):
        expected = ['timestamp', 'DestWeather', 'DistanceKilometers', 'AvgTicketPrice']

        field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME,
            display_names=expected
        )

        pd_flights = self.pd_flights()[expected]

        assert_series_equal(pd_flights.dtypes, field_mappings.dtypes())



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import numpy as np
import pandas as pd
from pandas.util.testing import assert_index_equal

# File called _pytest for PyCharm compatability
import eland as ed
from eland.tests import FLIGHTS_INDEX_NAME, ES_TEST_CLIENT
from eland.tests.common import TestData


class TestGetFieldNames(TestData):

    def test_get_field_names_all(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )
        pd_flights = self.pd_flights()

        fields1 = ed_field_mappings.get_field_names(include_scripted_fields=False)
        fields2 = ed_field_mappings.get_field_names(include_scripted_fields=True)

        assert fields1 == fields2
        assert_index_equal(pd_flights.columns, pd.Index(fields1))

    def test_get_field_names_selected(self):
        expected = ['Carrier', 'AvgTicketPrice']
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME,
            display_names=expected
        )
        pd_flights = self.pd_flights()[expected]

        fields1 = ed_field_mappings.get_field_names(include_scripted_fields=False)
        fields2 = ed_field_mappings.get_field_names(include_scripted_fields=True)

        assert fields1 == fields2
        assert_index_equal(pd_flights.columns, pd.Index(fields1))

    def test_get_field_names_scripted(self):
        expected = ['Carrier', 'AvgTicketPrice']
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME,
            display_names=expected
        )
        pd_flights = self.pd_flights()[expected]

        fields1 = ed_field_mappings.get_field_names(include_scripted_fields=False)
        fields2 = ed_field_mappings.get_field_names(include_scripted_fields=True)

        assert fields1 == fields2
        assert_index_equal(pd_flights.columns, pd.Index(fields1))

        # now add scripted field
        ed_field_mappings.add_scripted_field('scripted_field_None', None, np.dtype('int64'))

        fields3 = ed_field_mappings.get_field_names(include_scripted_fields=False)
        fields4 = ed_field_mappings.get_field_names(include_scripted_fields=True)

        assert fields1 == fields3
        fields1.append('scripted_field_None')
        assert fields1 == fields4



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# Originally based on code in pandasticsearch filters

# Es filter builder for BooleanCond
class BooleanFilter:
    def __init__(self, *args):
        self._filter = None

    def __and__(self, x):
        # Combine results
        if isinstance(self, AndFilter):
            if 'must_not' in x.subtree:
                # nest a must_not under a must
                self.subtree['must'].append(x.build())  # 'build includes bool'
            else:
                # append a must to a must
                self.subtree['must'].append(x.subtree)  # 'subtree strips bool'
            return self
        elif isinstance(x, AndFilter):
            if 'must_not' in self.subtree:
                x.subtree['must'].append(self.build())
            else:
                x.subtree['must'].append(self.subtree)
            return x
        return AndFilter(self, x)

    def __or__(self, x):
        # Combine results
        if isinstance(self, OrFilter):
            if 'must_not' in x.subtree:
                self.subtree['should'].append(x.build())
            else:
                self.subtree['should'].append(x.subtree)
            return self
        elif isinstance(x, OrFilter):
            if 'must_not' in self.subtree:
                x.subtree['should'].append(self.build())
            else:
                x.subtree['should'].append(self.subtree)
            return x
        return OrFilter(self, x)

    def __invert__(self):
        return NotFilter(self)

    def empty(self):
        if self._filter is None:
            return True
        return False

    def __repr__(self):
        return str(self._filter)

    @property
    def subtree(self):
        if 'bool' in self._filter:
            return self._filter['bool']
        else:
            return self._filter

    def build(self):
        return self._filter


# Binary operator
class AndFilter(BooleanFilter):
    def __init__(self, *args):
        [isinstance(x, BooleanFilter) for x in args]
        super(AndFilter, self).__init__()
        self._filter = {'bool': {'must': [x.build() for x in args]}}


class OrFilter(BooleanFilter):
    def __init__(self, *args):
        [isinstance(x, BooleanFilter) for x in args]
        super(OrFilter, self).__init__()
        self._filter = {'bool': {'should': [x.build() for x in args]}}


class NotFilter(BooleanFilter):
    def __init__(self, x):
        assert isinstance(x, BooleanFilter)
        super(NotFilter, self).__init__()
        self._filter = {'bool': {'must_not': x.build()}}


# LeafBooleanFilter
class GreaterEqual(BooleanFilter):
    def __init__(self, field, value):
        super(GreaterEqual, self).__init__()
        self._filter = {'range': {field: {'gte': value}}}


class Greater(BooleanFilter):
    def __init__(self, field, value):
        super(Greater, self).__init__()
        self._filter = {'range': {field: {'gt': value}}}


class LessEqual(BooleanFilter):
    def __init__(self, field, value):
        super(LessEqual, self).__init__()
        self._filter = {'range': {field: {'lte': value}}}


class Less(BooleanFilter):
    def __init__(self, field, value):
        super(Less, self).__init__()
        self._filter = {'range': {field: {'lt': value}}}


class Equal(BooleanFilter):
    def __init__(self, field, value):
        super(Equal, self).__init__()
        self._filter = {'term': {field: value}}


class IsIn(BooleanFilter):
    def __init__(self, field, value):
        super(IsIn, self).__init__()
        assert isinstance(value, list)
        if field == 'ids':
            self._filter = {'ids': {'values': value}}
        else:
            self._filter = {'terms': {field: value}}


class Like(BooleanFilter):
    def __init__(self, field, value):
        super(Like, self).__init__()
        self._filter = {'wildcard': {field: value}}


class Rlike(BooleanFilter):
    def __init__(self, field, value):
        super(Rlike, self).__init__()
        self._filter = {'regexp': {field: value}}


class Startswith(BooleanFilter):
    def __init__(self, field, value):
        super(Startswith, self).__init__()
        self._filter = {'prefix': {field: value}}


class IsNull(BooleanFilter):
    def __init__(self, field):
        super(IsNull, self).__init__()
        self._filter = {'missing': {'field': field}}


class NotNull(BooleanFilter):
    def __init__(self, field):
        super(NotNull, self).__init__()
        self._filter = {'exists': {'field': field}}


class ScriptFilter(BooleanFilter):
    def __init__(self, inline, lang=None, params=None):
        super(ScriptFilter, self).__init__()
        script = {'inline': inline}
        if lang is not None:
            script['lang'] = lang
        if params is not None:
            script['params'] = params
        self._filter = {'script': {'script': script}}



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
from datetime import datetime

import numpy as np
import pandas as pd
from pandas.util.testing import assert_series_equal

import eland as ed
from eland.tests.common import ES_TEST_CLIENT
from eland.tests.common import TestData
from eland.tests.common import assert_pandas_eland_frame_equal
from eland.tests.common import assert_pandas_eland_series_equal


class TestDataFrameDateTime(TestData):
    times = ["2019-11-26T19:58:15.246+0000",
             "1970-01-01T00:00:03.000+0000"]
    time_index_name = 'test_time_formats'

    @classmethod
    def setup_class(cls):
        """ setup any state specific to the execution of the given class (which
        usually contains tests).
        """
        es = ES_TEST_CLIENT
        if es.indices.exists(cls.time_index_name):
            es.indices.delete(index=cls.time_index_name)
        dts = [datetime.strptime(time, "%Y-%m-%dT%H:%M:%S.%f%z")
               for time in cls.times]

        time_formats_docs = [TestDataFrameDateTime.get_time_values_from_datetime(dt)
                             for dt in dts]
        mappings = {'properties': {}}

        for field_name, field_value in time_formats_docs[0].items():
            mappings['properties'][field_name] = {}
            mappings['properties'][field_name]['type'] = 'date'
            mappings['properties'][field_name]['format'] = field_name

        body = {"mappings": mappings}
        index = 'test_time_formats'
        es.indices.delete(index=index, ignore=[400, 404])
        es.indices.create(index=index, body=body)

        for i, time_formats in enumerate(time_formats_docs):
            es.index(index=index, body=time_formats, id=i)
        es.indices.refresh(index=index)

    @classmethod
    def teardown_class(cls):
        """ teardown any state that was previously setup with a call to
        setup_class.
        """

        es = ES_TEST_CLIENT
        es.indices.delete(index=cls.time_index_name)

    def test_datetime_to_ms(self):
        df = pd.DataFrame(data={'A': np.random.rand(3),
                                'B': 1,
                                'C': 'foo',
                                'D': pd.Timestamp('20190102'),
                                'E': [1.0, 2.0, 3.0],
                                'F': False,
                                'G': [1, 2, 3]},
                          index=['0', '1', '2'])

        expected_mappings = {'mappings': {
            'properties': {'A': {'type': 'double'},
                           'B': {'type': 'long'},
                           'C': {'type': 'keyword'},
                           'D': {'type': 'date'},
                           'E': {'type': 'double'},
                           'F': {'type': 'boolean'},
                           'G': {'type': 'long'}}}}

        mappings = ed.FieldMappings._generate_es_mappings(df)

        assert expected_mappings == mappings

        # Now create index
        index_name = 'eland_test_generate_es_mappings'

        ed_df = ed.pandas_to_eland(df, ES_TEST_CLIENT, index_name, es_if_exists="replace", es_refresh=True)

        #print(df.to_string())
        #print(ed_df.to_string())
        #print(ed_df.dtypes)
        #print(ed_df._to_pandas().dtypes)

        assert_series_equal(df.dtypes, ed_df.dtypes)

        assert_pandas_eland_frame_equal(df, ed_df)

    def test_all_formats(self):
        index_name = self.time_index_name
        ed_df = ed.read_es(ES_TEST_CLIENT, index_name)

        for format_name in self.time_formats.keys():
            times = [pd.to_datetime(datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S.%f%z")
                                    .strftime(self.time_formats[format_name]),
                                    format=self.time_formats[format_name])
                     for dt in self.times]

            ed_series = ed_df[format_name]
            pd_series = pd.Series(times,
                                  index=[str(i) for i in range(len(self.times))],
                                  name=format_name)

            assert_pandas_eland_series_equal(pd_series, ed_series)

    @staticmethod
    def get_time_values_from_datetime(dt: datetime) -> dict:
        time_formats = {
            "epoch_millis": int(dt.timestamp() * 1000),
            "epoch_second": int(dt.timestamp()),
            "strict_date_optional_time": dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "basic_date": dt.strftime("%Y%m%d"),
            "basic_date_time": dt.strftime("%Y%m%dT%H%M%S.%f")[:-3] + dt.strftime("%z"),
            "basic_date_time_no_millis": dt.strftime("%Y%m%dT%H%M%S%z"),
            "basic_ordinal_date": dt.strftime("%Y%j"),
            "basic_ordinal_date_time": dt.strftime("%Y%jT%H%M%S.%f")[:-3] + dt.strftime("%z"),
            "basic_ordinal_date_time_no_millis": dt.strftime("%Y%jT%H%M%S%z"),
            "basic_time": dt.strftime("%H%M%S.%f")[:-3] + dt.strftime("%z"),
            "basic_time_no_millis": dt.strftime("%H%M%S%z"),
            "basic_t_time": dt.strftime("T%H%M%S.%f")[:-3] + dt.strftime("%z"),
            "basic_t_time_no_millis": dt.strftime("T%H%M%S%z"),
            "basic_week_date": dt.strftime("%GW%V%u"),
            "basic_week_date_time": dt.strftime("%GW%V%uT%H%M%S.%f")[:-3] + dt.strftime("%z"),
            "basic_week_date_time_no_millis": dt.strftime("%GW%V%uT%H%M%S%z"),
            "strict_date": dt.strftime("%Y-%m-%d"),
            "date": dt.strftime("%Y-%m-%d"),
            "strict_date_hour": dt.strftime("%Y-%m-%dT%H"),
            "date_hour": dt.strftime("%Y-%m-%dT%H"),
            "strict_date_hour_minute": dt.strftime("%Y-%m-%dT%H:%M"),
            "date_hour_minute": dt.strftime("%Y-%m-%dT%H:%M"),
            "strict_date_hour_minute_second": dt.strftime("%Y-%m-%dT%H:%M:%S"),
            "date_hour_minute_second": dt.strftime("%Y-%m-%dT%H:%M:%S"),
            "strict_date_hour_minute_second_fraction": dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3],
            "date_hour_minute_second_fraction": dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3],
            "strict_date_hour_minute_second_millis": dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3],
            "date_hour_minute_second_millis": dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3],
            "strict_date_time": dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "date_time": dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "strict_date_time_no_millis": dt.strftime("%Y-%m-%dT%H:%M:%S%z"),
            "date_time_no_millis": dt.strftime("%Y-%m-%dT%H:%M:%S%z"),
            "strict_hour": dt.strftime("%H"),
            "hour": dt.strftime("%H"),
            "strict_hour_minute": dt.strftime("%H:%M"),
            "hour_minute": dt.strftime("%H:%M"),
            "strict_hour_minute_second": dt.strftime("%H:%M:%S"),
            "hour_minute_second": dt.strftime("%H:%M:%S"),
            "strict_hour_minute_second_fraction": dt.strftime("%H:%M:%S.%f")[:-3],
            "hour_minute_second_fraction": dt.strftime("%H:%M:%S.%f")[:-3],
            "strict_hour_minute_second_millis": dt.strftime("%H:%M:%S.%f")[:-3],
            "hour_minute_second_millis": dt.strftime("%H:%M:%S.%f")[:-3],
            "strict_ordinal_date": dt.strftime("%Y-%j"),
            "ordinal_date": dt.strftime("%Y-%j"),
            "strict_ordinal_date_time": dt.strftime("%Y-%jT%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "ordinal_date_time": dt.strftime("%Y-%jT%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "strict_ordinal_date_time_no_millis": dt.strftime("%Y-%jT%H:%M:%S%z"),
            "ordinal_date_time_no_millis": dt.strftime("%Y-%jT%H:%M:%S%z"),
            "strict_time": dt.strftime("%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "time": dt.strftime("%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "strict_time_no_millis": dt.strftime("%H:%M:%S%z"),
            "time_no_millis": dt.strftime("%H:%M:%S%z"),
            "strict_t_time": dt.strftime("T%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "t_time": dt.strftime("T%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "strict_t_time_no_millis": dt.strftime("T%H:%M:%S%z"),
            "t_time_no_millis": dt.strftime("T%H:%M:%S%z"),
            "strict_week_date": dt.strftime("%G-W%V-%u"),
            "week_date": dt.strftime("%G-W%V-%u"),
            "strict_week_date_time": dt.strftime("%G-W%V-%uT%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "week_date_time": dt.strftime("%G-W%V-%uT%H:%M:%S.%f")[:-3] + dt.strftime("%z"),
            "strict_week_date_time_no_millis": dt.strftime("%G-W%V-%uT%H:%M:%S%z"),
            "week_date_time_no_millis": dt.strftime("%G-W%V-%uT%H:%M:%S%z"),
            "strict_weekyear": dt.strftime("%G"),
            "weekyear": dt.strftime("%G"),
            "strict_weekyear_week": dt.strftime("%G-W%V"),
            "weekyear_week": dt.strftime("%G-W%V"),
            "strict_weekyear_week_day": dt.strftime("%G-W%V-%u"),
            "weekyear_week_day": dt.strftime("%G-W%V-%u"),
            "strict_year": dt.strftime("%Y"),
            "year": dt.strftime("%Y"),
            "strict_year_month": dt.strftime("%Y-%m"),
            "year_month": dt.strftime("%Y-%m"),
            "strict_year_month_day": dt.strftime("%Y-%m-%d"),
            "year_month_day": dt.strftime("%Y-%m-%d"),
        }

        return time_formats

    time_formats = {
        "epoch_millis": "%Y-%m-%dT%H:%M:%S.%f",
        "epoch_second": "%Y-%m-%dT%H:%M:%S",
        "strict_date_optional_time": "%Y-%m-%dT%H:%M:%S.%f%z",
        "basic_date": "%Y%m%d",
        "basic_date_time": "%Y%m%dT%H%M%S.%f",
        "basic_date_time_no_millis": "%Y%m%dT%H%M%S%z",
        "basic_ordinal_date": "%Y%j",
        "basic_ordinal_date_time": "%Y%jT%H%M%S.%f%z",
        "basic_ordinal_date_time_no_millis": "%Y%jT%H%M%S%z",
        "basic_time": "%H%M%S.%f%z",
        "basic_time_no_millis": "%H%M%S%z",
        "basic_t_time": "T%H%M%S.%f%z",
        "basic_t_time_no_millis": "T%H%M%S%z",
        "basic_week_date": "%GW%V%u",
        "basic_week_date_time": "%GW%V%uT%H%M%S.%f%z",
        "basic_week_date_time_no_millis": "%GW%V%uT%H%M%S%z",
        "date": "%Y-%m-%d",
        "strict_date": "%Y-%m-%d",
        "strict_date_hour": "%Y-%m-%dT%H",
        "date_hour": "%Y-%m-%dT%H",
        "strict_date_hour_minute": "%Y-%m-%dT%H:%M",
        "date_hour_minute": "%Y-%m-%dT%H:%M",
        "strict_date_hour_minute_second": "%Y-%m-%dT%H:%M:%S",
        "date_hour_minute_second": "%Y-%m-%dT%H:%M:%S",
        "strict_date_hour_minute_second_fraction": "%Y-%m-%dT%H:%M:%S.%f",
        "date_hour_minute_second_fraction": "%Y-%m-%dT%H:%M:%S.%f",
        "strict_date_hour_minute_second_millis": "%Y-%m-%dT%H:%M:%S.%f",
        "date_hour_minute_second_millis": "%Y-%m-%dT%H:%M:%S.%f",
        "strict_date_time": "%Y-%m-%dT%H:%M:%S.%f%z",
        "date_time": "%Y-%m-%dT%H:%M:%S.%f%z",
        "strict_date_time_no_millis": "%Y-%m-%dT%H:%M:%S%z",
        "date_time_no_millis": "%Y-%m-%dT%H:%M:%S%z",
        "strict_hour": "%H",
        "hour": "%H",
        "strict_hour_minute": "%H:%M",
        "hour_minute": "%H:%M",
        "strict_hour_minute_second": "%H:%M:%S",
        "hour_minute_second": "%H:%M:%S",
        "strict_hour_minute_second_fraction": "%H:%M:%S.%f",
        "hour_minute_second_fraction": "%H:%M:%S.%f",
        "strict_hour_minute_second_millis": "%H:%M:%S.%f",
        "hour_minute_second_millis": "%H:%M:%S.%f",
        "strict_ordinal_date": "%Y-%j",
        "ordinal_date": "%Y-%j",
        "strict_ordinal_date_time": "%Y-%jT%H:%M:%S.%f%z",
        "ordinal_date_time": "%Y-%jT%H:%M:%S.%f%z",
        "strict_ordinal_date_time_no_millis": "%Y-%jT%H:%M:%S%z",
        "ordinal_date_time_no_millis": "%Y-%jT%H:%M:%S%z",
        "strict_time": "%H:%M:%S.%f%z",
        "time": "%H:%M:%S.%f%z",
        "strict_time_no_millis": "%H:%M:%S%z",
        "time_no_millis": "%H:%M:%S%z",
        "strict_t_time": "T%H:%M:%S.%f%z",
        "t_time": "T%H:%M:%S.%f%z",
        "strict_t_time_no_millis": "T%H:%M:%S%z",
        "t_time_no_millis": "T%H:%M:%S%z",
        "strict_week_date": "%G-W%V-%u",
        "week_date": "%G-W%V-%u",
        "strict_week_date_time": "%G-W%V-%uT%H:%M:%S.%f%z",
        "week_date_time": "%G-W%V-%uT%H:%M:%S.%f%z",
        "strict_week_date_time_no_millis": "%G-W%V-%uT%H:%M:%S%z",
        "week_date_time_no_millis": "%G-W%V-%uT%H:%M:%S%z",
        "strict_weekyear_week_day": "%G-W%V-%u",
        "weekyear_week_day": "%G-W%V-%u",
        "strict_year": "%Y",
        "year": "%Y",
        "strict_year_month": "%Y-%m",
        "year_month": "%Y-%m",
        "strict_year_month_day": "%Y-%m-%d",
        "year_month_day": "%Y-%m-%d"
    }

    # excluding these formats as pandas throws a ValueError
    # "strict_weekyear": ("%G", None) - not supported in pandas
    # "strict_weekyear_week": ("%G-W%V", None),
    # E   ValueError: ISO year directive '%G' must be used with the ISO week directive '%V' and a weekday directive '%A', '%a', '%w', or '%u'.



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
import pytest
from pandas.util.testing import assert_series_equal

import eland as ed
from eland.tests import FLIGHTS_INDEX_NAME, FLIGHTS_MAPPING
from eland.tests.common import ES_TEST_CLIENT
from eland.tests.common import TestData


class TestFieldNamePDDType(TestData):

    def test_all_formats(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        pd_flights = self.pd_flights()

        assert_series_equal(pd_flights.dtypes, ed_field_mappings.dtypes())

        for es_field_name in FLIGHTS_MAPPING['mappings']['properties'].keys():
            pd_dtype = ed_field_mappings.field_name_pd_dtype(es_field_name)

            assert pd_flights[es_field_name].dtype == pd_dtype

    def test_non_existant(self):
        ed_field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        with pytest.raises(KeyError):
            pd_dtype = ed_field_mappings.field_name_pd_dtype('unknown')



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
import pytest

import eland as ed
from eland.tests import ES_TEST_CLIENT, FLIGHTS_INDEX_NAME
from eland.tests.common import TestData


class TestDisplayNames(TestData):

    def test_init_all_fields(self):
        field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        expected = self.pd_flights().columns.to_list()

        assert expected == field_mappings.display_names

    def test_init_selected_fields(self):
        expected = ['timestamp', 'DestWeather', 'DistanceKilometers', 'AvgTicketPrice']

        field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME,
            display_names=expected
        )

        assert expected == field_mappings.display_names

    def test_set_display_names(self):
        expected = ['Cancelled', 'timestamp', 'DestWeather', 'DistanceKilometers', 'AvgTicketPrice']

        field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        field_mappings.display_names = expected

        assert expected == field_mappings.display_names

        # now set again
        new_expected = ['AvgTicketPrice', 'timestamp']

        field_mappings.display_names = new_expected
        assert new_expected == field_mappings.display_names

    def test_not_found_display_names(self):
        not_found = ['Cancelled', 'timestamp', 'DestWeather', 'unknown', 'DistanceKilometers', 'AvgTicketPrice']

        field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        with pytest.raises(KeyError):
            field_mappings.display_names = not_found

        expected = self.pd_flights().columns.to_list()

        assert expected == field_mappings.display_names

    def test_invalid_list_type_display_names(self):
        field_mappings = ed.FieldMappings(
            client=ed.Client(ES_TEST_CLIENT),
            index_pattern=FLIGHTS_INDEX_NAME
        )

        # not a list like object
        with pytest.raises(ValueError):
            field_mappings.display_names = 12.0

        # tuple is list like
        field_mappings.display_names = ('Cancelled', 'DestWeather')

        expected = ['Cancelled', 'DestWeather']

        assert expected == field_mappings.display_names



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability

from eland.tests.common import TestData
from eland.tests.common import (
    assert_pandas_eland_frame_equal,
    assert_pandas_eland_series_equal
)


class TestDataFrameGetItem(TestData):

    def test_getitem_one_attribute(self):
        ed_flights = self.ed_flights().head(103)
        pd_flights = self.pd_flights().head(103)

        ed_flights_OriginAirportID = ed_flights['OriginAirportID']
        pd_flights_OriginAirportID = pd_flights['OriginAirportID']

        assert_pandas_eland_series_equal(pd_flights_OriginAirportID, ed_flights_OriginAirportID)

    def test_getitem_attribute_list(self):
        ed_flights = self.ed_flights().head(42)
        pd_flights = self.pd_flights().head(42)

        ed_flights_slice = ed_flights[['OriginAirportID', 'AvgTicketPrice', 'Carrier']]
        pd_flights_slice = pd_flights[['OriginAirportID', 'AvgTicketPrice', 'Carrier']]

        assert_pandas_eland_frame_equal(pd_flights_slice, ed_flights_slice)

    def test_getitem_one_argument(self):
        ed_flights = self.ed_flights().head(89)
        pd_flights = self.pd_flights().head(89)

        ed_flights_OriginAirportID = ed_flights.OriginAirportID
        pd_flights_OriginAirportID = pd_flights.OriginAirportID

        assert_pandas_eland_series_equal(pd_flights_OriginAirportID, ed_flights_OriginAirportID)

    def test_getitem_multiple_calls(self):
        ed_flights = self.ed_flights().head(89)
        pd_flights = self.pd_flights().head(89)

        ed_col0 = ed_flights[['DestCityName', 'DestCountry', 'DestLocation', 'DestRegion']]
        try:
            ed_col1 = ed_col0['Carrier']
        except KeyError:
            pass

        pd_col1 = pd_flights['DestCountry']
        ed_col1 = ed_col0['DestCountry']

        assert_pandas_eland_series_equal(pd_col1, ed_col1)



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
import numpy as np
import pytest

from eland.tests.common import TestData, assert_pandas_eland_series_equal


class TestSeriesArithmetics(TestData):

    def test_ecommerce_series_invalid_div(self):
        pd_df = self.pd_ecommerce()
        ed_df = self.ed_ecommerce()

        # eland / pandas == error
        with pytest.raises(TypeError):
            ed_series = ed_df['total_quantity'] / pd_df['taxful_total_price']

    def test_ecommerce_series_simple_arithmetics(self):
        pd_df = self.pd_ecommerce().head(100)
        ed_df = self.ed_ecommerce().head(100)

        pd_series = pd_df['taxful_total_price'] + 5 + pd_df['total_quantity'] / pd_df['taxless_total_price'] - pd_df[
            'total_unique_products'] * 10.0 + pd_df['total_quantity']
        ed_series = ed_df['taxful_total_price'] + 5 + ed_df['total_quantity'] / ed_df['taxless_total_price'] - ed_df[
            'total_unique_products'] * 10.0 + ed_df['total_quantity']

        assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

    def test_ecommerce_series_simple_integer_addition(self):
        pd_df = self.pd_ecommerce().head(100)
        ed_df = self.ed_ecommerce().head(100)

        pd_series = pd_df['taxful_total_price'] + 5
        ed_series = ed_df['taxful_total_price'] + 5

        assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

    def test_ecommerce_series_simple_series_addition(self):
        pd_df = self.pd_ecommerce().head(100)
        ed_df = self.ed_ecommerce().head(100)

        pd_series = pd_df['taxful_total_price'] + pd_df['total_quantity']
        ed_series = ed_df['taxful_total_price'] + ed_df['total_quantity']

        assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

    def test_ecommerce_series_basic_arithmetics(self):
        pd_df = self.pd_ecommerce().head(100)
        ed_df = self.ed_ecommerce().head(100)

        ops = ['__add__',
               '__truediv__',
               '__floordiv__',
               '__pow__',
               '__mod__',
               '__mul__',
               '__sub__',
               'add',
               'truediv',
               'floordiv',
               'pow',
               'mod',
               'mul',
               'sub']

        for op in ops:
            pd_series = getattr(pd_df['taxful_total_price'], op)(pd_df['total_quantity'])
            ed_series = getattr(ed_df['taxful_total_price'], op)(ed_df['total_quantity'])
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

            pd_series = getattr(pd_df['taxful_total_price'], op)(10.56)
            ed_series = getattr(ed_df['taxful_total_price'], op)(10.56)
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

            pd_series = getattr(pd_df['taxful_total_price'], op)(np.float32(1.879))
            ed_series = getattr(ed_df['taxful_total_price'], op)(np.float32(1.879))
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

            pd_series = getattr(pd_df['taxful_total_price'], op)(int(8))
            ed_series = getattr(ed_df['taxful_total_price'], op)(int(8))
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

    def test_supported_series_dtypes_ops(self):
        pd_df = self.pd_ecommerce().head(100)
        ed_df = self.ed_ecommerce().head(100)

        # Test some specific operations that are and aren't supported
        numeric_ops = ['__add__',
                       '__truediv__',
                       '__floordiv__',
                       '__pow__',
                       '__mod__',
                       '__mul__',
                       '__sub__']

        non_string_numeric_ops = ['__add__',
                                  '__truediv__',
                                  '__floordiv__',
                                  '__pow__',
                                  '__mod__',
                                  '__sub__']
        # __mul__ is supported for int * str in pandas

        # float op float
        for op in numeric_ops:
            pd_series = getattr(pd_df['taxful_total_price'], op)(pd_df['taxless_total_price'])
            ed_series = getattr(ed_df['taxful_total_price'], op)(ed_df['taxless_total_price'])
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

        # int op float
        for op in numeric_ops:
            pd_series = getattr(pd_df['total_quantity'], op)(pd_df['taxless_total_price'])
            ed_series = getattr(ed_df['total_quantity'], op)(ed_df['taxless_total_price'])
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

        # float op int
        for op in numeric_ops:
            pd_series = getattr(pd_df['taxful_total_price'], op)(pd_df['total_quantity'])
            ed_series = getattr(ed_df['taxful_total_price'], op)(ed_df['total_quantity'])
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

        # str op int (throws)
        for op in non_string_numeric_ops:
            with pytest.raises(TypeError):
                pd_series = getattr(pd_df['currency'], op)(pd_df['total_quantity'])
            with pytest.raises(TypeError):
                ed_series = getattr(ed_df['currency'], op)(ed_df['total_quantity'])
            with pytest.raises(TypeError):
                pd_series = getattr(pd_df['currency'], op)(1)
            with pytest.raises(TypeError):
                ed_series = getattr(ed_df['currency'], op)(1)

        # int op str (throws)
        for op in non_string_numeric_ops:
            with pytest.raises(TypeError):
                pd_series = getattr(pd_df['total_quantity'], op)(pd_df['currency'])
            with pytest.raises(TypeError):
                ed_series = getattr(ed_df['total_quantity'], op)(ed_df['currency'])

    def test_ecommerce_series_basic_rarithmetics(self):
        pd_df = self.pd_ecommerce().head(10)
        ed_df = self.ed_ecommerce().head(10)

        ops = ['__radd__',
               '__rtruediv__',
               '__rfloordiv__',
               '__rpow__',
               '__rmod__',
               '__rmul__',
               '__rsub__',
               'radd',
               'rtruediv',
               'rfloordiv',
               'rpow',
               'rmod',
               'rmul',
               'rsub']

        for op in ops:
            pd_series = getattr(pd_df['taxful_total_price'], op)(pd_df['total_quantity'])
            ed_series = getattr(ed_df['taxful_total_price'], op)(ed_df['total_quantity'])
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

            pd_series = getattr(pd_df['taxful_total_price'], op)(3.141)
            ed_series = getattr(ed_df['taxful_total_price'], op)(3.141)
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

            pd_series = getattr(pd_df['taxful_total_price'], op)(np.float32(2.879))
            ed_series = getattr(ed_df['taxful_total_price'], op)(np.float32(2.879))
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

            pd_series = getattr(pd_df['taxful_total_price'], op)(int(6))
            ed_series = getattr(ed_df['taxful_total_price'], op)(int(6))
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

    def test_supported_series_dtypes_rops(self):
        pd_df = self.pd_ecommerce().head(100)
        ed_df = self.ed_ecommerce().head(100)

        # Test some specific operations that are and aren't supported
        numeric_ops = ['__radd__',
                       '__rtruediv__',
                       '__rfloordiv__',
                       '__rpow__',
                       '__rmod__',
                       '__rmul__',
                       '__rsub__']

        non_string_numeric_ops = ['__radd__',
                                  '__rtruediv__',
                                  '__rfloordiv__',
                                  '__rpow__',
                                  '__rmod__',
                                  '__rsub__']
        # __rmul__ is supported for int * str in pandas

        # float op float
        for op in numeric_ops:
            pd_series = getattr(pd_df['taxful_total_price'], op)(pd_df['taxless_total_price'])
            ed_series = getattr(ed_df['taxful_total_price'], op)(ed_df['taxless_total_price'])
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

        # int op float
        for op in numeric_ops:
            pd_series = getattr(pd_df['total_quantity'], op)(pd_df['taxless_total_price'])
            ed_series = getattr(ed_df['total_quantity'], op)(ed_df['taxless_total_price'])
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

        # float op int
        for op in numeric_ops:
            pd_series = getattr(pd_df['taxful_total_price'], op)(pd_df['total_quantity'])
            ed_series = getattr(ed_df['taxful_total_price'], op)(ed_df['total_quantity'])
            assert_pandas_eland_series_equal(pd_series, ed_series, check_less_precise=True)

        # str op int (throws)
        for op in non_string_numeric_ops:
            with pytest.raises(TypeError):
                pd_series = getattr(pd_df['currency'], op)(pd_df['total_quantity'])
            with pytest.raises(TypeError):
                ed_series = getattr(ed_df['currency'], op)(ed_df['total_quantity'])
            with pytest.raises(TypeError):
                pd_series = getattr(pd_df['currency'], op)(10.0)
            with pytest.raises(TypeError):
                ed_series = getattr(ed_df['currency'], op)(10.0)

        # int op str (throws)
        for op in non_string_numeric_ops:
            with pytest.raises(TypeError):
                pd_series = getattr(pd_df['total_quantity'], op)(pd_df['currency'])
            with pytest.raises(TypeError):
                ed_series = getattr(ed_df['total_quantity'], op)(ed_df['currency'])



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

import sys
from abc import ABC, abstractmethod

from eland import QueryCompiler

"""
NDFrame
---------
Abstract base class for eland.DataFrame and eland.Series.

The underlying data resides in Elasticsearch and the API aligns as much as
possible with pandas APIs.

This allows the eland.DataFrame to access large datasets stored in Elasticsearch,
without storing the dataset in local memory.

Implementation Details
----------------------

Elasticsearch indexes can be configured in many different ways, and these indexes
utilise different data structures to pandas.

eland.DataFrame operations that return individual rows (e.g. df.head()) return
_source data. If _source is not enabled, this data is not accessible.

Similarly, only Elasticsearch searchable fields can be searched or filtered, and
only Elasticsearch aggregatable fields can be aggregated or grouped.

"""


class NDFrame(ABC):

    def __init__(self,
                 client=None,
                 index_pattern=None,
                 columns=None,
                 index_field=None,
                 query_compiler=None):
        """
        pandas.DataFrame/Series like API that proxies into Elasticsearch index(es).

        Parameters
        ----------
        client : eland.Client
            A reference to a Elasticsearch python client
        """
        if query_compiler is None:
            query_compiler = QueryCompiler(client=client, index_pattern=index_pattern, display_names=columns,
                                           index_field=index_field)
        self._query_compiler = query_compiler

    def _get_index(self):
        """
        Return eland index referencing Elasticsearch field to index a DataFrame/Series

        Returns
        -------
        eland.Index:
            Note eland.Index has a very limited API compared to pandas.Index

        See Also
        --------
        :pandas_api_docs:`pandas.DataFrame.index`
        :pandas_api_docs:`pandas.Series.index`

        Examples
        --------
        >>> df = ed.DataFrame('localhost', 'flights')
        >>> assert isinstance(df.index, ed.Index)
        >>> df.index.index_field
        '_id'
        >>> s = df['Carrier']
        >>> assert isinstance(s.index, ed.Index)
        >>> s.index.index_field
        '_id'
        """
        return self._query_compiler.index

    index = property(_get_index)

    @property
    def dtypes(self):
        """
        Return the pandas dtypes in the DataFrame. Elasticsearch types are mapped
        to pandas dtypes via Mappings._es_dtype_to_pd_dtype.__doc__

        Returns
        -------
        pandas.Series
            The data type of each column.

        See Also
        --------
        :pandas_api_docs:`pandas.DataFrame.dtypes`

        Examples
        --------
        >>> df = ed.DataFrame('localhost', 'flights', columns=['Origin', 'AvgTicketPrice', 'timestamp', 'dayOfWeek'])
        >>> df.dtypes
        Origin                    object
        AvgTicketPrice           float64
        timestamp         datetime64[ns]
        dayOfWeek                  int64
        dtype: object
        """
        return self._query_compiler.dtypes

    def _build_repr(self, num_rows):
        # self could be Series or DataFrame
        if len(self.index) <= num_rows:
            return self._to_pandas()

        num_rows = num_rows

        head_rows = int(num_rows / 2) + num_rows % 2
        tail_rows = num_rows - head_rows

        head = self.head(head_rows)._to_pandas()
        tail = self.tail(tail_rows)._to_pandas()

        return head.append(tail)

    def __sizeof__(self):
        # Don't default to pandas, just return approximation TODO - make this more accurate
        return sys.getsizeof(self._query_compiler)

    def __len__(self):
        """Gets the length of the DataFrame.

        Returns:
            Returns an integer length of the DataFrame object.
        """
        return len(self.index)

    def _info_es(self, buf):
        self._query_compiler.info_es(buf)

    def mean(self, numeric_only=True):
        """
        Return mean value for each numeric column

        TODO - implement remainder of pandas arguments, currently non-numerics are not supported

        Returns
        -------
        pandas.Series
            mean value for each numeric column

        See Also
        --------
        :pandas_api_docs:`pandas.DataFrame.mean`

        Examples
        --------
        >>> df = ed.DataFrame('localhost', 'flights')
        >>> df.mean()
        AvgTicketPrice         628.253689
        Cancelled                0.128494
        DistanceKilometers    7092.142457
        DistanceMiles         4406.853010
        FlightDelay              0.251168
        FlightDelayMin          47.335171
        FlightTimeHour           8.518797
        FlightTimeMin          511.127842
        dayOfWeek                2.835975
        dtype: float64
        """
        return self._query_compiler.mean(numeric_only=numeric_only)

    def sum(self, numeric_only=True):
        """
        Return sum for each numeric column

        TODO - implement remainder of pandas arguments, currently non-numerics are not supported

        Returns
        -------
        pandas.Series
            sum for each numeric column

        See Also
        --------
        :pandas_api_docs:`pandas.DataFrame.sum`

        Examples
        --------
        >>> df = ed.DataFrame('localhost', 'flights')
        >>> df.sum()
        AvgTicketPrice        8.204365e+06
        Cancelled             1.678000e+03
        DistanceKilometers    9.261629e+07
        DistanceMiles         5.754909e+07
        FlightDelay           3.280000e+03
        FlightDelayMin        6.181500e+05
        FlightTimeHour        1.112470e+05
        FlightTimeMin         6.674818e+06
        dayOfWeek             3.703500e+04
        dtype: float64
        """
        return self._query_compiler.sum(numeric_only=numeric_only)

    def min(self, numeric_only=True):
        """
        Return the minimum value for each numeric column

        TODO - implement remainder of pandas arguments, currently non-numerics are not supported

        Returns
        -------
        pandas.Series
            min value for each numeric column

        See Also
        --------
        :pandas_api_docs:`pandas.DataFrame.min`

        Examples
        --------
        >>> df = ed.DataFrame('localhost', 'flights')
        >>> df.min()
        AvgTicketPrice        100.020531
        Cancelled               0.000000
        DistanceKilometers      0.000000
        DistanceMiles           0.000000
        FlightDelay             0.000000
        FlightDelayMin          0.000000
        FlightTimeHour          0.000000
        FlightTimeMin           0.000000
        dayOfWeek               0.000000
        dtype: float64
        """
        return self._query_compiler.min(numeric_only=numeric_only)

    def max(self, numeric_only=True):
        """
        Return the maximum value for each numeric column

        TODO - implement remainder of pandas arguments, currently non-numerics are not supported

        Returns
        -------
        pandas.Series
            max value for each numeric column

        See Also
        --------
        :pandas_api_docs:`pandas.DataFrame.max`

        Examples
        --------
        >>> df = ed.DataFrame('localhost', 'flights')
        >>> df.max()
        AvgTicketPrice         1199.729004
        Cancelled                 1.000000
        DistanceKilometers    19881.482422
        DistanceMiles         12353.780273
        FlightDelay               1.000000
        FlightDelayMin          360.000000
        FlightTimeHour           31.715034
        FlightTimeMin          1902.901978
        dayOfWeek                 6.000000
        dtype: float64
        """
        return self._query_compiler.max(numeric_only=numeric_only)

    def nunique(self):
        """
        Return cardinality of each field.

        **Note we can only do this for aggregatable Elasticsearch fields - (in general) numeric and keyword
        rather than text fields**

        This method will try and field aggregatable fields if possible if mapping has::

            "customer_first_name" : {
              "type" : "text",
              "fields" : {
                "keyword" : {
                  "type" : "keyword",
                  "ignore_above" : 256
                }
              }
            }

        we will aggregate ``customer_first_name`` columns using ``customer_first_name.keyword``.

        TODO - implement remainder of pandas arguments

        Returns
        -------
        pandas.Series
            cardinality of each column

        See Also
        --------
        :pandas_api_docs:`pandas.DataFrame.nunique`

        Examples
        --------
        >>> columns = ['category', 'currency', 'customer_birth_date', 'customer_first_name', 'user']
        >>> df = ed.DataFrame('localhost', 'ecommerce', columns=columns)
        >>> df.nunique()
        category                6
        currency                1
        customer_birth_date     0
        customer_first_name    46
        user                   46
        dtype: int64
        """
        return self._query_compiler.nunique()

    def _hist(self, num_bins):
        return self._query_compiler._hist(num_bins)

    def describe(self):
        """
        Generate descriptive statistics that summarize the central tendency, dispersion and shape of a
        dataset’s distribution, excluding NaN values.

        Analyzes both numeric and object series, as well as DataFrame column sets of mixed data types.
        The output will vary depending on what is provided. Refer to the notes below for more detail.

        TODO - add additional arguments (current only numeric values supported)

        Returns
        -------
        pandas.Dataframe:
            Summary information

        See Also
        --------
        :pandas_api_docs:`pandas.DataFrame.describe`

        Examples
        --------
        >>> df = ed.DataFrame('localhost', 'flights', columns=['AvgTicketPrice', 'FlightDelayMin'])
        >>> df.describe() # ignoring percentiles as they don't generate consistent results
               AvgTicketPrice  FlightDelayMin
        count    13059.000000    13059.000000
        mean       628.253689       47.335171
        std        266.386661       96.743006
        min        100.020531        0.000000
        ...
        ...
        ...
        max       1199.729004      360.000000
        """
        return self._query_compiler.describe()

    @abstractmethod
    def _to_pandas(self, show_progress=False):
        pass

    @abstractmethod
    def head(self, n=5):
        pass

    @abstractmethod
    def tail(self, n=5):
        pass



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

from eland.filter import *


class TestOperators:
    def test_leaf_boolean_filter(self):
        assert GreaterEqual('a', 2).build() == {"range": {"a": {"gte": 2}}}
        assert LessEqual('a', 2).build() == {"range": {"a": {"lte": 2}}}
        assert Less('a', 2).build() == {"range": {"a": {"lt": 2}}}
        assert Equal('a', 2).build() == {"term": {"a": 2}}
        exp = Equal('a', 2)
        assert (~exp).build()['bool'], {"must_not": {"term": {"a": 2}}}
        assert Greater('a', 2).build() == {"range": {"a": {"gt": 2}}}
        assert IsIn('a', [1, 2, 3]).build() == {'terms': {'a': [1, 2, 3]}}
        assert Like('a', 'a*b').build() == {'wildcard': {'a': 'a*b'}}
        assert Rlike('a', 'a*b').build() == {'regexp': {'a': 'a*b'}}
        assert Startswith('a', 'jj').build() == {'prefix': {'a': 'jj'}}
        assert IsNull('a').build() == {'missing': {'field': 'a'}}
        assert NotNull('a').build() == {'exists': {'field': 'a'}}
        assert ScriptFilter('doc["num1"].value > params.param1', params={'param1': 5}).build() == {
            'script': {'script': {'inline': 'doc["num1"].value > params.param1', 'params': {'param1': 5}}}}
        assert IsIn('ids', [1, 2, 3]).build() == {'ids': {'values': [1, 2, 3]}}

    def test_and_filter1(self):
        exp = GreaterEqual('a', 2) & Less('b', 3)
        assert exp.build() == {'bool': {'must': [{'range': {'a': {'gte': 2}}}, {'range': {'b': {'lt': 3}}}]}}

    def test_and_filter2(self):
        exp = GreaterEqual('a', 2) & Less('b', 3) & Equal('c', 4)
        assert exp.build() == \
               {
                   'bool': {
                       'must': [
                           {'range': {'a': {'gte': 2}}},
                           {'range': {'b': {'lt': 3}}},
                           {'term': {'c': 4}}
                       ]
                   }
               }

    def test_and_filter3(self):
        exp = GreaterEqual('a', 2) & (Less('b', 3) & Equal('c', 4))
        assert exp.build() == \
               {
                   'bool': {
                       'must': [
                           {'range': {'b': {'lt': 3}}},
                           {'term': {'c': 4}},
                           {'range': {'a': {'gte': 2}}}
                       ]
                   }
               }

    def test_or_filter1(self):
        exp = GreaterEqual('a', 2) | Less('b', 3)
        assert exp.build() == \
               {
                   'bool': {
                       'should': [
                           {'range': {'a': {'gte': 2}}},
                           {'range': {'b': {'lt': 3}}}
                       ]
                   }
               }

    def test_or_filter2(self):
        exp = GreaterEqual('a', 2) | Less('b', 3) | Equal('c', 4)
        assert exp.build() == \
               {
                   'bool': {
                       'should': [
                           {'range': {'a': {'gte': 2}}},
                           {'range': {'b': {'lt': 3}}},
                           {'term': {'c': 4}}
                       ]
                   }
               }

    def test_or_filter3(self):
        exp = GreaterEqual('a', 2) | (Less('b', 3) | Equal('c', 4))
        assert exp.build() == \
               {
                   'bool': {
                       'should': [
                           {'range': {'b': {'lt': 3}}},
                           {'term': {'c': 4}},
                           {'range': {'a': {'gte': 2}}}
                       ]
                   }
               }

    def test_not_filter(self):
        exp = ~GreaterEqual('a', 2)
        assert exp.build() == \
               {
                   'bool': {
                       'must_not': {'range': {'a': {'gte': 2}}}
                   }
               }

    def test_not_not_filter(self):
        exp = ~~GreaterEqual('a', 2)

        assert exp.build() == \
               {
                   'bool': {
                       'must_not': {
                           'bool': {
                               'must_not': {'range': {'a': {'gte': 2}}}
                           }
                       }
                   }
               }

    def test_not_and_filter(self):
        exp = ~(GreaterEqual('a', 2) & Less('b', 3))
        assert exp.build() == \
               {
                   'bool': {
                       'must_not': {
                           'bool': {
                               'must': [
                                   {'range': {'a': {'gte': 2}}},
                                   {'range': {'b': {'lt': 3}}}
                               ]
                           }
                       }
                   }
               }

    def test_and_or_filter(self):
        exp = GreaterEqual('a', 2) & (Less('b', 3) | Equal('c', 4))
        assert exp.build() == \
               {
                   'bool': {
                       'must': [
                           {'range': {'a': {'gte': 2}}},
                           {
                               'bool': {
                                   'should': [
                                       {'range': {'b': {'lt': 3}}},
                                       {'term': {'c': 4}}
                                   ]
                               }
                           }
                       ]
                   }
               }

    def test_and_not_or_filter(self):
        exp = GreaterEqual('a', 2) & ~(Less('b', 3) | Equal('c', 4))
        assert exp.build() == \
               {
                   'bool': {
                       'must': [
                           {'range': {'a': {'gte': 2}}},
                           {
                               'bool': {
                                   'must_not': {
                                       'bool': {
                                           'should': [
                                               {'range': {'b': {'lt': 3}}},
                                               {'term': {'c': 4}}
                                           ]
                                       }

                                   }
                               }
                           }
                       ]
                   }
               }

    def test_must_and_must_not_filter(self):
        exp = (GreaterEqual('a', 2) & GreaterEqual('b', 2)) & ~(IsIn('ids', [1, 2, 3]))
        a = exp.build()
        b = {
            'bool': {
                'must': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'gte': 2}}},
                    {
                        'bool': {
                            'must_not': {
                                'ids': {'values': [1, 2, 3]}
                            }
                        }
                    }
                ]
            }
        }
        assert a == b

    def test_must_not_and_must_filter(self):
        exp = ~(IsIn('ids', [1, 2, 3])) & (GreaterEqual('a', 2) & GreaterEqual('b', 2))
        a = exp.build()
        b = {
            'bool': {
                'must': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'gte': 2}}},
                    {
                        'bool': {
                            'must_not': {
                                'ids': {'values': [1, 2, 3]}
                            }
                        }
                    }
                ]
            }
        }
        assert a == b

    def test_must_not_or_must_filter(self):
        exp = ~(IsIn('ids', [1, 2, 3])) | (GreaterEqual('a', 2) | GreaterEqual('b', 2))
        a = exp.build()
        b = {
            'bool': {
                'should': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'gte': 2}}},
                    {
                        'bool': {
                            'must_not': {
                                'ids': {'values': [1, 2, 3]}
                            }
                        }
                    }
                ]
            }
        }
        assert a == b

    def test_must_or_must_not_filter(self):
        exp = (GreaterEqual('a', 2) | GreaterEqual('b', 2)) | ~(IsIn('ids', [1, 2, 3]))
        a = exp.build()
        b = {
            'bool': {
                'should': [
                    {'range': {'a': {'gte': 2}}},
                    {'range': {'b': {'gte': 2}}},
                    {
                        'bool': {
                            'must_not': {
                                'ids': {'values': [1, 2, 3]}
                            }
                        }
                    }
                ]
            }
        }
        assert a == b



#  Copyright 2019 Elasticsearch BV
#
#      Licensed under the Apache License, Version 2.0 (the "License");
#      you may not use this file except in compliance with the License.
#      You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0
#
#      Unless required by applicable law or agreed to in writing, software
#      distributed under the License is distributed on an "AS IS" BASIS,
#      WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#      See the License for the specific language governing permissions and
#      limitations under the License.

# File called _pytest for PyCharm compatability
import numpy as np

from eland.tests.common import TestData
from eland.tests.common import (
    assert_pandas_eland_frame_equal
)


class TestDataFrameSelectDTypes(TestData):

    def test_select_dtypes_include_number(self):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_flights_numeric = ed_flights.select_dtypes(include=[np.number])
        pd_flights_numeric = pd_flights.select_dtypes(include=[np.number])

        assert_pandas_eland_frame_equal(pd_flights_numeric.head(103), ed_flights_numeric.head(103))

    def test_select_dtypes_exclude_number(self):
        ed_flights = self.ed_flights()
        pd_flights = self.pd_flights()

        ed_flights_non_numeric = ed_flights.select_dtypes(exclude=[np.number])
        pd_flights_non_numeric = pd_flights.select_dtypes(exclude=[np.number])

        assert_pandas_eland_frame_equal(pd_flights_non_numeric.head(103), ed_flights_non_numeric.head(103))
